###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import UfoBase, ValueType, RetainedFactory, GetVal
from onyx.core import FloatField, ReferenceField, StringField, DictField

__all__ = ["Exchange"]


###############################################################################
class Exchange(UfoBase):
    """
    Class used to represent an exchange or trading venue.
    """
    IsoCode = StringField()
    Denominated = ReferenceField(obj_type="Currency")
    Multiplier = FloatField(default=1.0)
    HolidayCalendar = ReferenceField(obj_type="HolidayCalendar")
    Country = ReferenceField(obj_type="Category")
    Region = ReferenceField(obj_type="Category")
    FullName = StringField(default="")
    Codes = DictField()
    TimeZone = StringField()
    MarketSession = StringField()

    # -------------------------------------------------------------------------
    def __post_init__(self):
        if self.IsoCode is None:
            raise ValueError("IsoCode not set")
        self.Name = self.Name or "{0:s} Exchange".format(self.IsoCode)
        if self.HolidayCalendar is None:
            self.HolidayCalendar = GetVal(self.Denominated, "HolidayCalendar")

    # -------------------------------------------------------------------------
    @RetainedFactory()
    def Platform(self, graph):
        return "Bloomberg"

    # -------------------------------------------------------------------------
    @ValueType()
    def Code(self, graph):
        return graph(self, "Codes")[graph(self, "Platform")]


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing
    from . import ufo_currency
    from . import ufo_database
    from . import ufo_holiday_calendar
    from . import ufo_category

    ufo_database.prepare_for_test()
    ufo_holiday_calendar.prepare_for_test()
    ufo_currency.prepare_for_test()
    ufo_category.prepare_for_test()

    exchanges = [
        {
            "IsoCode": "XNYS",
            "Denominated": "USD",
            "Country": "USA",
            "Region": "Nth. America",
            "FullName": "New York Stock Exchange",
            "Codes": {
                "Bloomberg": "US",
                "Google": "NYSE",
                "Yahoo": "",
                "WSJ": "US",
            },
        },
        {
            "IsoCode": "XLON",
            "Denominated": "GBP",
            "Country": "United Kingdom",
            "Region": "Nth. Europe",
            "FullName": "London Stock Exchange",
            "Codes": {
                "Bloomberg": "LN",
                "Google": "LON",
                "Yahoo": "L",
                "WSJ": "UK",
            },
        },
        {
            "IsoCode": "XEUR",
            "Denominated": "EUR",
            "Country": "Germany",
            "Region": "Cnt. Europe",
            "FullName": "Eurex",
            "Codes": {"Bloomberg": "EUX"},
        },
        {
            "IsoCode": "IFEN",
            "Denominated": "EUR",
            "Country": "United Kingdom",
            "Region": "Nth. Europe",
            "FullName": "ICE Futures Europe",
            "Codes": {"Bloomberg": "ICE"},
        }
    ]

    for info in exchanges:
        AddIfMissing(Exchange(**info))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import ReferenceField, GetVal

from ..system.ufo_forward_cash import ForwardCash


###############################################################################
class CashDividend(ForwardCash):
    """
    Tradable class that represents a cash dividend payable/receivable.
    """
    Asset = ReferenceField(obj_type="EquityAsset")

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        sym = GetVal(self.Asset, "Symbol")
        code = GetVal(GetVal(self.Asset, "Exchange"), "Code")
        pay_date = self.PaymentDate.strftime("%d%m%y")
        return ("DVD {0:s} {1:2s} {2:3s} "
                "{3:6s}").format(sym, code, self.Currency, pay_date)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Structure, DateOffset
from onyx.core import ValueType, GetVal
from onyx.core import FloatField, StringField, ReferenceField

from .ufo_commod_forward import CommodForward

from ..system.tradable_api import AgingTradableObj, AddByInference
from ..system.tradable_api import HashStoredAttrs
from ..system.ufo_forward_cash import ForwardCash

__all__ = ["CommodFutures"]


###############################################################################
class CommodFutures(AgingTradableObj):
    """
    Tradable class that represents a commodity futures contract with physical
    delivery.
    """
    Asset = ReferenceField(obj_type="CommodCnt")
    FixedPrice = FloatField(positive=True)

    # --- the settlement currency can differ from the denomination of the asset
    SettlementCcy = ReferenceField(obj_type="Currency")

    # --- fx rate uesd to convert the fixed price (which is expressed in the
    #     currency of the asset) to the settlement currency
    SettlementFx = FloatField()

    # --- the date rule for the cash payment, defaults to 2 biz days after
    #     settlement
    PaymentDateRule = StringField("+2b")

    # -------------------------------------------------------------------------
    def __post_init__(self):
        asset_ccy = GetVal(self.Asset, "Denominated")

        self.SettlementCcy = self.SettlementCcy or asset_ccy

        if self.SettlementFx is None:
            # --- default to spot
            if asset_ccy == self.SettlementCcy:
                fx = 1.0
            else:
                asset_cross = "{0:3s}/USD".format(asset_ccy)
                sett_cross = "{0:3s}/USD".format(self.SettlementCcy)
                fx = GetVal(asset_cross, "Spot") / GetVal(sett_cross, "Spot")

            self.SettlementFx = fx

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        asset = graph(self, "Asset")
        mul = graph(asset, "Multiplier")
        lot = graph(asset, "ContractSize")
        cal = graph(asset, "HolidayCalendar")

        sett_fx = graph(self, "SettlementFx")
        sett_date = graph(asset, "FutSettDate")
        pay_rule = graph(self, "PaymentDateRule")

        fwd_info = {
            "Asset": graph(asset, "CommodAsset"),
            "AvgStartDate": sett_date,
            "AvgEndDate": sett_date,
            "AvgType": "LAST",
            "Margined": True,
        }
        cash_info = {
            "Currency": graph(self, "SettlementCcy"),
            "PaymentDate": DateOffset(sett_date, pay_rule, cal),
        }

        fwd_sec = AddByInference(CommodForward(**fwd_info), True)
        cash_sec = AddByInference(ForwardCash(**cash_info), True)

        # --- use list-based constructor to enforce key ordering
        return Structure([
            (fwd_sec.Name,  lot),
            (cash_sec.Name, -lot*mul*sett_fx*graph(self, "FixedPrice"))
        ])

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm = 0.0
        for sec, qty in graph(self, "Leaves").items():
            mtm += qty*graph(sec, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        sett_ccy = graph(self, "SettlementCcy")
        spot_fx = graph("{0:3s}/USD".format(sett_ccy), "Spot")
        return graph(self, "MktValUSD") / spot_fx

    # -------------------------------------------------------------------------
    @ValueType()
    def ExpirationDate(self, graph):
        asset = graph(self, "Asset")
        cnt = graph(asset, "GetContract", graph(self, "DeliveryMonth"))
        return graph(cnt, "FutSettDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def NextTransactionDate(self, graph):
        return graph(self, "ExpirationDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def TradeTypes(self, graph):
        mapping = super().TradeTypes
        mapping.update({
            "Delivery": "DeliverySecurities",
        })
        return mapping

    # -------------------------------------------------------------------------
    @ValueType()
    def ExpectedTransaction(self, graph):
        return "Delivery"

    # -------------------------------------------------------------------------
    @ValueType()
    def DeliverySecurities(self, graph):
        """
        Delivery of physical commodity: here we only include the cash leg
        """
        asset = graph(self, "Asset")
        sett_ccy = graph(self, "SettlementCcy")
        sett_date = graph(asset, "FutSettDate")
        pay_rule = graph(self, "PaymentDateRule")
        cal = graph(asset, "HolidayCalendar")

        mul = graph(asset, "Multiplier")
        lot = graph(asset, "ContractSize")

        cash = ForwardCash(Currency=sett_ccy,
                           PaymentDate=DateOffset(sett_date, pay_rule, cal))

        return [
            {"Security": AddByInference(cash, in_memory=True),
             "Quantity": -mul*lot*graph(self, "FixedPrice")},
        ]

    # -------------------------------------------------------------------------
    #  this value type is used by edit screens
    @ValueType()
    def AssetSpot(self, graph):
        return graph(graph(self, "Asset"), "Spot")

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        mkt = GetVal(self.Asset, "Market")
        sym = GetVal(self.Asset, "Symbol")
        cnt = GetVal(self.Asset, "DeliveryMonth")
        mush = HashStoredAttrs(self, 3)
        return ("CmdFUT {0:s} {1:s} {2:3s}"
                " {3:3s} {{0:2d}}").format(mkt, sym, cnt, mush)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from . import ufo_commod_contract
    from ..system import ufo_forward_cash

    cnt, prc = ufo_commod_contract.prepare_for_test()[0]
    ufo_forward_cash.prepare_for_test()

    info = {
        "Asset": cnt,
        "FixedPrice": prc - 0.1,
    }

    securities = [AddByInference(CommodFutures(**info))]

    return [sec.Name for sec in securities]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from numpy import where, vectorize
from numpy import log as np_log, sqrt as np_sqrt, exp as np_exp
from math import log as m_log, sqrt as m_sqrt
from math import exp as m_exp, expm1 as m_expm1, erf as m_erf, pi as m_pi

from scipy.optimize import bisect

__all__ = ["opt_BS", "opt_BAW", "opt_JZ", "quick_delta", "implied_vol"]

# --- local variables and functions
invsqrt2 = 1.0 / m_sqrt(2.0)

m_ncdf = lambda x: 0.5*(1.0 + m_erf(x*invsqrt2))
v_ncdf = vectorize(m_ncdf, otypes=[float])


# -----------------------------------------------------------------------------
def opt_BS(opt_type, spot, strike, vol, term, rd, rf, ret_type="PREMIUM"):
    """
    Description:
        Black-Scholes option price and greeks for different european option
        types.
        Black76 formula for futures is retrived by setting rd = rf = r.
    Inputs:
        opt_type - option type (CALL, PUT). Only first letter is used.
        spot     - current spot price
        strike   - strike of the option
        vol      - price volatility (standard deviation of log returns)
        term     - time to expiration (units must be consistent with what is
                   used for vol, rd, rf)
        rd       - domestic risk-free interest rate
        rf       - foreign risk-free interest rate (or dividend yield)
        ret_type - quantity to be computed (PREMIUM, DELTA). Only first letter
                   is used.
    Returns:
        A float as per ret_type.
    """
    # --- convert string arguments to upper-case
    opt_type = opt_type[0].upper()
    ret_type = ret_type[0].upper()

    # --- precompute as much as possible
    volsqrtt = vol*np_sqrt(term)

    # --- early return if possible
    if volsqrtt <= 0.00025:
        # --- either volatility is zero or very small or we are extremely close
        #     to expiration. Return based on intrinsic value.

        # --- Calls
        if opt_type == "C":
            if ret_type == "P":
                moneyness = spot*np_exp((rd - rf)*term) - strike
                return m_exp(-rd*term)*where(moneyness <= 0.0, 0.0, moneyness)
            elif ret_type == "D":
                otm = spot*np_exp((rd - rf)*term) <= strike
                return where(otm, 0.0, np_exp(-rf*term))
            else:
                NameError("Unrecognized return type, {0:s}".format(ret_type))

        # --- Puts
        elif opt_type == "P":
            if ret_type == "P":
                moneyness = strike - spot*np_exp((rd - rf)*term)
                return m_exp(-rd*term)*where(moneyness <= 0.0, 0.0, moneyness)
            elif ret_type == "D":
                otm = strike <= spot*np_exp((rd - rf)*term)
                return where(otm, 0.0, -np_exp(-rf*term))
            else:
                NameError("Unrecognized return type, {0:s}".format(ret_type))
        else:
            raise NameError("Unrecognized option type, {0:s}".format(opt_type))

    # --- calculate according to Black-Scholes
    d1 = (np_log(spot / strike) + (rd - rf + 0.5*vol*vol)*term) / volsqrtt

    # --- Calls
    if opt_type == "C":
        if ret_type == "P":
            return (np_exp(-rf*term)*spot*v_ncdf(d1) -
                    np_exp(-rd*term)*strike*v_ncdf(d1 - volsqrtt))
        elif ret_type == "D":
            return np_exp(-rf*term)*v_ncdf(d1)
        else:
            NameError("Unrecognized return type, {0:s}".format(ret_type))

    # --- Puts
    elif opt_type == "P":
        if ret_type == "P":
            return (np_exp(-rd*term)*strike*v_ncdf(volsqrtt - d1) -
                    np_exp(-rf*term)*spot*v_ncdf(-d1))
        elif ret_type == "D":
            return -np_exp(-rf*term)*v_ncdf(-d1)
        else:
            NameError("Unrecognized return type, {0:s}".format(ret_type))

    else:
        raise NameError("Unrecognized option type, {0:s}".format(opt_type))


# -----------------------------------------------------------------------------
def opt_BAW(opt_type, spot, strike, vol, term, rd, rf):
    """
    Description:
        Barone-Adesi Whaley approximation for american options (premium only).
        The formula for futures is retrived by setting rd = rf = r.
    Inputs:
        opt_type - option type (CALL, PUT). Only first letter is used.
        spot     - current spot price
        strike   - strike of the option
        vol      - price volatility (standard deviation of log returns)
        term     - time to expiration (units must be consistent with what is
                   used for vol, rd, rf)
        rd       - domestic risk-free interest rate (cannot be negative)
        rf       - foreign risk-free interest rate (or dividend yield)
    Returns:
        A float.
    """
    if rd < 0.0:
        raise ValueError("Barone-Adesi Whaley approximation "
                         "doesn't work for negative interest rates")

    # --- convert string arguments to upper-case
    opt_type = opt_type[0].upper()

    # --- precompute as much as possible
    volsq = vol*vol

    if term == 0.0 or volsq <= 0.00005:
        disc = m_exp(-rd*term)
        if opt_type == "C":
            return max(max(spot - strike, 0.0),
                       max(spot*m_exp((rd - rf)*term) - strike, 0.0)*disc)
        elif opt_type == "P":
            return max(max(strike - spot, 0.0),
                       max(strike - spot*m_exp((rd - rf)*term), 0.0)*disc)
        else:
            raise NameError("Unrecognized option type, {0:s}".format(opt_type))

    # --- this is actually -(beta-1) using Hull's notation
    b = 1.0 - 2.0*(rd - rf) / volsq

    if opt_type == "C":
        if rd > 0.0:
            h = - m_expm1(-rd*term)
            a = 2.0*rd / volsq
            g = 0.5*(b + m_sqrt(b*b + 4.0*a / h))
        else:
            g = 0.5*(b + m_sqrt(b*b + 8.0 / (term*volsq)))

        ss = critical_price_BAW(opt_type, strike, vol, term, rd, rf, g)

        if spot < ss:
            # --- below critical price: option should not be exercised
            d1 = ((m_log(ss / strike) +
                  (rd - rf + 0.5*volsq)*term) / (vol*m_sqrt(term)))
            A = ss / g * (1.0 - m_exp(-rf*term)*m_ncdf(d1))

            return (A*(spot / ss)**g +
                    opt_BS("C", spot, strike, vol, term, rd, rf))
        else:
            # --- option should be exercised
            return spot - strike

    elif opt_type == "P":
        if rd > 0.0:
            h = - m_expm1(-rd*term)
            a = 2.0*rd / volsq
            g = 0.5*(b - m_sqrt(b*b + 4.0*a / h))
        else:
            g = 0.5*(b - m_sqrt(b*b + 8.0 / (term*volsq)))

        ss = critical_price_BAW(opt_type, strike, vol, term, rd, rf, g)

        if spot > ss:
            # --- above critical price: option should not be exercised
            d1 = ((m_log(ss / strike) +
                  (rd - rf + 0.5*volsq)*term) / (vol*m_sqrt(term)))
            A = - ss / g * (1.0 - m_exp(-rf*term)*m_ncdf(-d1))

            return (A*(spot / ss)**g +
                    opt_BS("P", spot, strike, vol, term, rd, rf))
        else:
            # --- option should be exercised
            return strike - spot

    else:
        raise NameError("Unrecognized option type, {0:s}".format(opt_type))


# -----------------------------------------------------------------------------
def critical_price_BAW(opt_type, K, vol, term, rd, rf, g):
    """
    Auxiliary function used by the Barone-Adesi Whaley approximation for
    american options.
    """
    # --- pre-calculate constant terms
    drift = (rd - rf + 0.5*vol*vol)*term
    volsqrtt = vol*m_sqrt(term)
    invden = 1.0 / volsqrtt
    disc_rd = m_exp(-rd*term)
    disc_rf = m_exp(-rf*term)

    if opt_type == "C":
        def error(ss):
            d1 = (m_log(ss / K) + drift)*invden
            bs = disc_rf*ss*m_ncdf(d1) - disc_rd*K*m_ncdf(d1 - volsqrtt)
            return bs + (1.0 - disc_rf*m_ncdf(d1)) * ss / g - (ss - K)

        # --- early return if possible
        if error(2.0*K) > 0.0:
            return 2.0*K
        else:
            return bisect(error, K, 2.0*K, xtol=0.0005, full_output=False)

    else:
        def error(ss):
            d1 = (m_log(ss / K) + drift)*invden
            bs = disc_rd*K*m_ncdf(volsqrtt - d1) - disc_rf*ss*m_ncdf(-d1)
            return bs - (1.0 - disc_rf*m_ncdf(-d1)) * ss / g - (K - ss)

        return bisect(error, 0.0005, K, xtol=0.0005, full_output=False)


# -----------------------------------------------------------------------------
def opt_JZ(opt_type, spot, strike, vol, term, rd, rf):
    """
    Description:
        Ju-Zhong approximation for american options (premium only).
        The formula for futures is retrived by setting rd = rf = r.
    Inputs:
        opt_type - option type (CALL, PUT). Only first letter is used.
        spot     - current spot price
        strike   - strike of the option
        vol      - price volatility (standard deviation of log returns)
        term     - time to expiration (units must be consistent with what is
                   used for vol, rd, rf)
        rd       - domestic risk-free interest rate (cannot be negative)
        rf       - foreign risk-free interest rate (or dividend yield)
    Returns:
        A float.
    """
    if rd <= 0.0:
        raise ValueError("Ju-Zhong approximation doesn't "
                         "work for zero or negative interest rates")

    # --- convert string arguments to upper-case
    opt_type = opt_type[0].upper()

    # --- precompute as much as possible
    volsq = vol*vol

    if term == 0.0 or volsq <= 0.00005:
        disc = m_exp(-rd*term)
        if opt_type == "C":
            return max(max(spot - strike, 0.0),
                       max(spot*m_exp((rd - rf)*term) - strike, 0.0)*disc)
        elif opt_type == "P":
            return max(max(strike - spot, 0.0),
                       max(strike - spot*m_exp((rd - rf)*term), 0.0)*disc)
        else:
            raise NameError("Unrecognized option type, {0:s}".format(opt_type))

    h = - m_expm1(-rd*term)
    a = 2.0*rd / volsq
    # --- this is actually -(beta-1) using Hull's notation
    b = 1.0 - 2.0*(rd - rf) / volsq

    if opt_type == "C":
        g = 0.5*(b + m_sqrt(b*b + 4.0*a / h))
        ss = critical_price_BAW(opt_type, strike, vol, term, rd, rf, g)

        if spot < ss:
            # --- below critical price: option should not be exercised
            rdf = m_exp(-(rd - rf)*term) / rd
            vsqt = vol*m_sqrt(term)

            hA = ss - strike - opt_BS("C", ss, strike, vol, term, rd, rf)
            d1 = (m_log(ss / strike) + (rd - rf + 0.5*volsq)*term) / vsqt

            gp = - a / (h*h*m_sqrt(b*b + 4.0*a / h))
            dV = (strike*m_ncdf(d1 - vsqt) - rf*ss*m_ncdf(d1)*rdf +
                  0.5*ss*m_exp(-0.5*d1*d1)*volsq*rdf / vsqt / m_sqrt(2.0*m_pi))

            dn = 1.0 / (2.0*g - b)
            c1 = 0.5*(1.0 - h)*a*gp*dn
            c2 = -(1.0 - h)*a*dn*(dV / hA + 1.0 / h + gp*dn)

            lns = m_log(spot / ss)
            chi = c1*lns*lns + c2*lns

            return (hA*(spot / ss)**g / (1.0 - chi) +
                    opt_BS("C", spot, strike, vol, term, rd, rf))
        else:
            # --- option should be exercised
            return spot - strike

    elif opt_type == "P":
        g = 0.5*(b - m_sqrt(b*b + 4.0*a / h))
        ss = critical_price_BAW(opt_type, strike, vol, term, rd, rf, g)

        if spot > ss:
            # --- above critical price: option should not be exercised
            rdf = m_exp(-(rd - rf)*term) / rd
            vsqt = vol*m_sqrt(term)

            hA = strike - ss - opt_BS("P", ss, strike, vol, term, rd, rf)
            d1 = (m_log(ss / strike) + (rd - rf + 0.5*volsq)*term) / vsqt

            gp = a / (h*h*m_sqrt(b*b + 4.0*a / h))
            dV = (rf*ss*m_ncdf(-d1)*rdf - strike*m_ncdf(vsqt - d1) +
                  0.5*ss*m_exp(-0.5*d1*d1)*volsq*rdf / vsqt / m_sqrt(2.0*m_pi))

            dn = 1.0 / (2.0*g - b)
            c1 = 0.5*(1.0 - h)*a*gp*dn
            c2 = -(1.0 - h)*a*dn*(dV / hA + 1.0 / h + gp*dn)

            lns = m_log(spot / ss)
            chi = c1*lns*lns + c2*lns

            return (hA*(spot / ss)**g / (1.0 - chi) +
                    opt_BS("P", spot, strike, vol, term, rd, rf))
        else:
            # --- option should be exercised
            return strike - spot

    else:
        raise NameError("Unrecognized option type, {0:s}".format(opt_type))


# -----------------------------------------------------------------------------
def quick_delta(spot, strike, vol, term):
    """
    Description:
        Quick-Delta is a measure of moneyness (from 0 to 1) and is an
        approximation to the delta of a put option.
    Inputs:
        spot   - current spot price
        strike - strike of the option
        vol    - price volatility (stanard deviation of log returns)
        term   - time to expiration (units must be consistent with vol, rd, rf)
    """
    if term > 0.0:
        return m_ncdf(np_log(strike / spot) / (vol*np_sqrt(term)))
    else:
        return 1.0 if spot < strike else 0.0


# -----------------------------------------------------------------------------
def implied_vol(opt_type, premium, spot, strike, term, rd, rf,
                american=False, pricer=None, vol_min=0.0, vol_max=2.5):
    """
    Description:
        Computes the implied Black-Scholes volatility given option parameters
        and price.
        For american options, by default it uses the Barone-Adesi Whaley
        approximation (different pricers can be specified).
    Inputs:
        opt_type - option type (CALL, PUT). Only first letter is used.
        premium  - the option premium
        spot     - current spot price
        strike   - strike of the option
        term     - time to expiration (units must be consistent with what is
                   used for vol, rd, rf)
        rd       - domestic risk-free interest rate
        rf       - foreign risk-free interest rate (or dividend yield)
        american - if True, it uses the BAW pricer (unless a different pricer
                   is indicated)
        pricer   - the pricer to use for american options
        vol_min  - min volatility allowed (used for iterative root finding)
        vol_max  - max volatility allowed (used for iterative root finding)
    Returns:
        The implied volatility
    """
    if american:
        if pricer is None:
            pricer = opt_BAW

        def error(vol):
            return premium - pricer(opt_type, spot, strike, vol, term, rd, rf)

        if error(vol_min) == 0.0:
            # --- no extrinsic value: IV is defined as the largest vol such
            #     that the option still has no extrinsic value
            def error(vol):
                return (1.e-6 + premium -
                        pricer(opt_type, spot, strike, vol, term, rd, rf))

    else:
        def error(vol):
            return premium - opt_BS(opt_type, spot, strike, vol, term, rd, rf)

    return bisect(error, vol_min, vol_max, xtol=0.0005, full_output=False)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import numpy as np
import math

__all__ = ["RegressionError", "DemingRegression", "RobustRegression"]


###############################################################################
class RegressionError(Exception):
    pass


# -----------------------------------------------------------------------------
def DemingRegression(x, y):
    """
    Description:
        Deming regression tries to find the line of best fit for a two
        dimensional dataset. It differs from the simple linear regression in
        that it accounts for errors in observations on both the x- and the
        y- axis.
        See: http://en.wikipedia.org/wiki/Deming_regression
    Inputs:
        x - dependent input vector
        y - independent input vector
    Returns:
        A tuple containing:
            - slope
            - intercept
            - signal/noise ratio defined as the ratio between the variances
              along the two pricipal components
            - vector of residuals
    """
    if len(x) < 3 or len(y) < 3:
        raise RegressionError("Insuffcient data for regression")

    xmean = x.mean()
    ymean = y.mean()

    xm = x - xmean
    ym = y - ymean

    den = 1.0 / (len(x) - 1.0)
    sxx = np.dot(xm, xm)*den
    syy = np.dot(ym, ym)*den
    sxy = np.dot(xm, ym)*den

    m = 0.5*(syy - sxx + math.sqrt((syy - sxx)**2 + 4.0*sxy*sxy)) / sxy
    c = ymean - m*xmean

    # --- rotate all point so that the regression line corresponds to the
    #     abscissas and the residuals to the ordinates (i.e. use pricipal
    #     components)
    theta = -math.atan(m)
    cos_t = math.cos(theta)
    sin_t = math.sin(theta)

    rx = (x - c)*cos_t - y*sin_t
    ry = (x - c)*sin_t + y*cos_t

    return m, c, rx.var() / ry.var(), ry - ry.mean()


# -----------------------------------------------------------------------------
def RobustRegression(x, y, thr=2.0):
    """
    Description:
        Two steps linear fit (symmetric in x and y), based on DemingRegression:
            - first identify best fit using the whole sample
            - remove outliers (see thr below) and refit model
    Inputs:
        x   - independent input vector
        y   - dependent input vector
        thr - only include points with residual distance (in units of std) from
              the mean of less than this value.
    Returns:
        A tuple, see DemingRegression
    """
    m, c, _, res = DemingRegression(x, y)

    # --- NB: residuals have zero mean by construction
    idx = np.nonzero(np.ravel(np.abs(res) < thr*res.std()))

    m, c, snr, res = DemingRegression(x[idx], y[idx])

    # --- residuals are calculated including all datapoints
    res = (y - m*x - c) * math.cos(-math.atan(m))

    return m, c, snr, res



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################
"""
Core functions for multi asset two fctors Gabillon Model (G2F):

     dF(t,T)            -beta*(T-t)                      -beta*(T-t)
    -------- = sigma_S*e          *dW_S + sigma_L*( 1 - e           )*dW_L
     F(t,T)

    where: <dW_S,dW_L> = rho_SL*dt
"""

from math import exp, log, sqrt
import numpy as np

__all__ = [
    "LocalVarG2F",
    "TermVarG2F",
    "TermCovG2F",
    "TermCovG2F2Assets",
    "G2FCalibrator",
    "G2FFuturesCalibrator",
    "SwapVolG2F",
]


# -----------------------------------------------------------------------------
def MrFactor(tau, beta, T):
    """
    Description:
        Mean-reversion function B:

                         tau
                        /       - beta*( T - t )
            B(tau,T) =  | dt * e
                        /
                        0
    """
    return exp(-beta*T)*(exp(beta*tau) - 1.0) / beta


# -----------------------------------------------------------------------------
def MrFactorSq(tau, beta_k, beta_l, T_i, T_j):
    """
    Description:
        Mean-reversion function B2:

                          tau
                         /       - beta_k*( T_i - t ) - beta_l*( T_j - t )
            B2(tau,T) =  | dt * e
                         /
                         0
    """
    return (exp(-beta_k*T_i - beta_l*T_j) *
            (exp((beta_k + beta_l)*tau) - 1.0) / (beta_k + beta_l))


# -----------------------------------------------------------------------------
# Cross Correlations between diagonal terms

def rho_YL(rho_SL_kl, rho_SL_k, rho_SL_k_comp, rho_LL_kl):
    return (rho_SL_kl - rho_SL_k*rho_LL_kl) / rho_SL_k_comp


def rho_YY(rho_SL_k, rho_SL_k_comp, rho_SL_l, rho_SL_l_comp,
           rho_SS_kl, rho_LL_kl, rho_YL_kl, rho_YL_lk):
    """
    rho_SL_k, rho_SL_k_comp:  G2F parameters for asset k
    rho_SL_l, rho_SL_l_comp:  G2F parameters for asset l
    rho_SS_kl, rho_LL_kl, rho_YL_kl, rho_YL_lk:  cross terms k-l
    """
    den = rho_SL_k_comp * rho_SL_l_comp
    num = (rho_SS_kl -
           rho_SL_k*rho_SL_l*rho_LL_kl -
           rho_SL_k*rho_SL_l_comp*rho_YL_lk -
           rho_SL_l*rho_SL_k_comp*rho_YL_kl)
    return num / den


# -----------------------------------------------------------------------------
def LocalVarG2F(sig_S, sig_L, rho_SL, beta, dT):
    """
    Two factors local variance
    """
    rho_SL_cmop_sq = 1.0 - rho_SL*rho_SL

    exp1 = exp(-beta*dT)
    exp2 = exp(-2.0*beta*dT)

    return (sig_S*sig_S*rho_SL_cmop_sq*exp2 +
            (sig_S*rho_SL*exp1 + sig_L*(1.0 - exp1))**2)


# -----------------------------------------------------------------------------
def TermVarG2F(sig_S, sig_L, rho_SL, beta, tau, T):
    """
    Description:
        G2F: terminal variance for a two factors model

                               tau
                          1   /   dF(t,T)   dF(t,T)
            var(tau,T) = ---  | < ------- * ------- >
                         tau  /    F(t,T)    F(t,T)
                              0
    Inputs:
        sig_S  -  short factor volatility
        sig_L  -  long factor volatility
        rho_SL -  short-long correlation
        beta   -  mean-reversion factor
        tau    -  time to option's expiration
        T      -  time to expiration of futures contract F
    """
    B2 = MrFactorSq(tau, beta, beta, T, T)
    B = MrFactor(tau, beta, T)

    return (sig_L*sig_L*tau - 2.0*B*(sig_L*sig_L - rho_SL*sig_S*sig_L) +
            B2*(sig_S*sig_S + sig_L*sig_L - 2.0*rho_SL*sig_S*sig_L)) / tau


# -----------------------------------------------------------------------------
def TermCovG2F(sig_S, sig_L, rho_SL, beta, tau, T_i, T_j):
    """
    Description:
        G2F: terminal covariance between two futures contracts F(t,T_i) and
             F(t,T_j) on the same asset.

                               tau
                          1   /   dF(t,T_i)   dF(t,T_j)
            var(tau,T) = ---  | < --------- * --------- >
                         tau  /    F(t,T_i)    F(t,T_j)
                              0
    Inputs:
        sig_S  -  short factor volatility
        sig_L  -  long factor volatility
        rho_SL -  short-long correlation
        beta   -  mean-reversion factor
        tau    -  time to option's expiration
        T_i    -  time to expiration of futures contract F_k(T_i)
        T_j    -  time to expiration of futures contract F_l(T_j)
    """
    B2 = MrFactorSq(tau, beta, beta, T_i, T_j)
    Bij = MrFactor(tau, beta, T_i) + MrFactor(tau, beta, T_j)

    return (sig_L*sig_L*tau - Bij*(sig_L*sig_L - rho_SL*sig_S*sig_L) +
            B2*(sig_S*sig_S + sig_L*sig_L - 2.0*rho_SL*sig_S*sig_L)) / tau


# -----------------------------------------------------------------------------
def TermCovG2F2Assets(sig_S_k, sig_L_k, rho_SL_k, rho_SL_k_comp, beta_k,
                      sig_S_l, sig_L_l, rho_SL_l, rho_SL_l_comp, beta_l,
                      rho_YY_kl, rho_YL_kl, rho_YL_lk, rho_LL_kl,
                      tau, T_i, T_j):
    """
    Signature:
        TermCovG2F( sig_S_k, sig_L_k, rho_SL_k, rho_SL_k_comp, beta_k,
                    sig_S_l, sig_L_l, rho_SL_l, rho_SL_l_comp, beta_l,
                    rho_YY_kl, rho_YL_kl, rho_YL_lk, rho_LL_kl,
                    tau, T_i, T_j )
    Description:
        G2F: terminal covariance between two futures contracts F_k(t,T_i) and
             F_l(t,T_j) on different assets (k and l).

                                 tau
                            1   /   dF_k(t,T_i)   dF_l(t,T_j)
        cov(tau,T_i,T_j) = ---  | < ----------- * ----------- >
                           tau  /    F_k(t,T_i)    F_l(t,T_j)
                                0
    Inputs:
        sig_S_k   -  short factor volatility for asset k
        sig_L_k   -  long factor volatility for asset k
        rho_SL_k  -  short-long correlation for asset k
        beta_k    -  mean-reversion factor for asset k
        sig_S_l   -  short factor volatility for asset l
        sig_L_l   -  long factor volatility for asset l
        rho_SL_l  -  short-long correlation for asset l
        beta_l    -  mean-reversion factor for asset l
        rho_YY_kl -  cross-asset YY correlation (as returned by rho_YY)
        rho_YL_kl -  cross-asset YL correlation (as returned by rho_YL)
        rho_YL_lk -  cross-asset LY correlation (as returned by rho_YL)
        rho_LL_kl -  cross-asset LL correlation
        tau       -  time to option's expiration
        T_i       -  time to expiration of futures contract F_k(T_i)
        T_j       -  time to expiration of futures contract F_l(T_j)
    """
    sig_hat_SL_kl = sig_S_k*sig_L_l*rho_SL_k
    sig_hat_SL_lk = sig_S_l*sig_L_k*rho_SL_l
    sig_hat_SL_kl_comp = sig_S_k*sig_L_l*rho_SL_k_comp
    sig_hat_SL_lk_comp = sig_S_l*sig_L_k*rho_SL_l_comp

    sig_hat_LL_kl = sig_L_k*sig_L_l*rho_LL_kl

    sig_hat_kl = (sig_S_k*sig_S_l*(rho_SL_k_comp*rho_SL_l_comp*rho_YY_kl +
                                   rho_SL_k_comp*rho_YL_kl*rho_SL_l +
                                   rho_SL_l_comp*rho_YL_lk*rho_SL_k +
                                   rho_SL_k*rho_SL_l*rho_LL_kl) -
                  sig_hat_SL_kl_comp*rho_YL_kl - sig_hat_SL_lk_comp*rho_YL_lk -
                  rho_LL_kl*(sig_hat_SL_kl + sig_hat_SL_lk) + sig_hat_LL_kl)

    sig_hat_k = (sig_hat_SL_kl_comp*rho_YL_kl +
                 sig_hat_SL_kl*rho_LL_kl - sig_hat_LL_kl)
    sig_hat_l = (sig_hat_SL_lk_comp*rho_YL_lk +
                 sig_hat_SL_lk*rho_LL_kl - sig_hat_LL_kl)

    B2_klij = MrFactorSq(tau, beta_k, beta_l, T_i, T_j)
    B_ki = MrFactor(tau, beta_k, T_i)
    B_lj = MrFactor(tau, beta_l, T_j)

    return (sig_hat_kl*B2_klij +
            sig_hat_k*B_ki + sig_hat_l*B_lj + sig_hat_LL_kl*tau) / tau


# -----------------------------------------------------------------------------
class G2FCalibError(Exception):
    pass


# -----------------------------------------------------------------------------
def G2FCalibrator(sig_Black, sig_L, rho_SL, beta, tau, T):
    """
    Description:
        G2F: calibrate sig_S to term volatilities marked by traders
    Inputs:
        sig_Black -  term (Black) volatility
        sig_L     -  long factor volatility
        rho_SL    -  short-long correlation
        beta      -  mean-reversion factor
        tau       -  time to option's expiration
        T         -  time to expiration of futures contract F
    """
    B2 = MrFactorSq(tau, beta, beta, T, T)
    B = MrFactor(tau, beta, T)
    a = B2
    b = sig_L*rho_SL*(B - B2)
    c = sig_L*sig_L*(B2 - 2.0*B + tau) - sig_Black*sig_Black*tau

    delta = b*b - a*c

    if delta >= 0:
        return (-b + sqrt(delta)) / a
    else:
        raise G2FCalibError("Cannot calibrate short volatility: delta < 0")


# -----------------------------------------------------------------------------
def G2FFuturesCalibrator(sig_Black, sig_L, rho_SL, beta, tau, T):
    """
    Description:
        G2F: calibrate sig_S to term volatilities marked by traders
    Inputs:
        sig_Black -  term (Black) volatility
        sig_L     -  long factor volatility (as percentage of short vol)
        rho_SL    -  short-long correlation
        beta      -  mean-reversion factor
        tau       -  time to option's expiration
        T         -  time to expiration of futures contract F
    """
    B2 = MrFactorSq(tau, beta, beta, T, T)
    B = MrFactor(tau, beta, T)
    a = ((1.0 + sig_L*sig_L - 2.0*sig_L*rho_SL)*B2 -
         2.0*sig_L*(sig_L - rho_SL)*B + sig_L*sig_L*tau)

    return sig_Black*sqrt(tau / a)


# -----------------------------------------------------------------------------
def SwapVolG2F(PD, ED, VEDs, sig_S, sig_L, rho_SL, beta, F=None):
    """
    Description:
        This is the Black volatility for a swaption. The underlying swap is
        defined by the volatility end dates (VEDs) of the monthly contracts.
    Inputs:
        PD     - Pricing Date
        ED     - Option's expiration date
        VEDs   - Volatility End Dates (i.e. last trading day for the
                 underlying monthly contract)
        sig_S  - Sigma Short (G2F)
        sig_L  - Sigma Long (G2F)
        rho_SL - Short-Long correlation (G2F)
        beta   - Beta (G2F)
        F      - Forward curve for the underlying monthly contracts
    """
    n = len(VEDs)
    tau = (ED.ordinal - PD.ordinal) / 365.0

    # --- early return for expired swaptions
    if tau <= 0.0:
        return 0.0

    # --- if no forward prices are passed, assume they are all equal to one
    if F is None:
        F = np.ones(n)

    w = np.ones(n) / n
    T = [(ved.ordinal - PD.ordinal) / 365.0 for ved in VEDs]

    CovMat = np.zeros((n, n))

    sig_B2 = (sig_S*sig_S + sig_L*sig_L - 2.0*sig_S*sig_L*rho_SL)
    sig_B1 = (sig_L*sig_L - sig_S*sig_L*rho_SL)
    sig_Lsq = sig_L*sig_L

    for i in range(n):
        Bi = exp(-beta*T[i])*(exp(beta*tau) - 1.0) / beta
        for j in range(i, n):
            Bj = exp(-beta*T[j])*(exp(beta*tau) - 1.0) / beta
            B2 = 0.5*exp(-beta*(T[i] + T[j]))*(exp(2.0*beta*tau) - 1.0) / beta
            cov = (sig_B2*B2 - sig_B1*(Bi + Bj) + sig_Lsq*tau) / tau

            CovMat[i,j] = cov  # analysis:ignore
            CovMat[j,i] = cov  # analysis:ignore

    fw = F*w
    M1 = np.sum(fw)
    M2 = np.dot(np.dot(fw.T, np.exp(CovMat)), fw)

    return sqrt(log(M2 / (M1*M1)))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import ValueType, ReferenceField

from .tradable_api import TradableObj

__all__ = ["CashBalance"]


###############################################################################
class CashBalance(TradableObj):
    """
    Tradable class that represents a cash amount in a given currrency.
    """
    Currency = ReferenceField(obj_type="Currency")

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        return 1.0

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        return graph("{0:3s}/USD".format(graph(self, "Currency")), "Spot")

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        return "CASH {0:3s}".format(self.Currency)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from .tradable_api import AddByInference
    from . import ufo_database
    from . import ufo_holiday_calendar
    from . import ufo_currency
    from . import ufo_currency_cross

    ufo_database.prepare_for_test()
    ufo_holiday_calendar.prepare_for_test()
    ufo_currency.prepare_for_test()
    ufo_currency_cross.prepare_for_test()

    securities = []
    for ccy in ["USD", "EUR", "GBP"]:
        securities.append(AddByInference(CashBalance(Currency=ccy)))

    return [sec.Name for sec in securities]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetVal, ChildrenSet, OnyxTestCase
from onyx.equity import ufo_equity_index_futures, ufo_equity_index_contract

import unittest


###############################################################################
class UnitTest(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.cnt, _ = ufo_equity_index_contract.prepare_for_test()[0]
        self.sec = ufo_equity_index_futures.prepare_for_test()[0]

    # -------------------------------------------------------------------------
    def test_MktVal(self):
        self.assertAlmostEqual(GetVal(self.sec, "MktVal"), 1000, 8)

    # -------------------------------------------------------------------------
    def test_MktValUSD(self):
        self.assertAlmostEqual(GetVal(self.sec, "MktValUSD"), 1150, 8)

    # -------------------------------------------------------------------------
    def test_TradeTypes(self):
        tt = {
            "Buy": "BuySecurities",
            "Sell": "SellSecurities",
            "Settlement": "SettlementSecurities",
        }
        self.assertEqual(GetVal(self.sec, "TradeTypes"), tt)

    # -------------------------------------------------------------------------
    def test_Children(self):
        kids = ChildrenSet((self.sec, "MktValUSD"), "Spot")
        self.assertEqual(kids, {self.cnt, "EUR/USD"})


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Curve, HlocvCurve, AddObj, GetObj, DelObj
from onyx.core import ObjExists, ObjNotFound, ObjDbTransaction
from onyx.core import TsDbGetCurve, TsNotFound
from onyx.core import GetVal, ValueType, FloatField
from onyx.core import DictField, SetField, StringField, ReferenceField

from ..system.ufo_asset import WithAssetAttributes
from .ufo_equity_index_contract import EquityIndexCnt

__all__ = ["EquityIndex", "MissingContract"]


###############################################################################
class MissingContract(Exception):
    pass


###############################################################################
class EquityIndex(WithAssetAttributes):
    """
    class used to represent an equity index and to provide access to futures
    contracts and options that settle on that index.
    """
    Tickers = DictField()
    ContractSize = FloatField()
    Country = ReferenceField(obj_type="Category")
    Region = ReferenceField(obj_type="Category")
    Sector = StringField(default="Equity Index")
    Subsector = StringField(default="Equity Index")
    TimeSeries = StringField()

    # --- settlement date rule for the underlying futures contracts
    SettDateRule = StringField()

    # --- date rule for the expiry of options
    OptExpDateRule = StringField()

    # --- the set of futures contracts expressed in LYY format
    Contracts = SetField(default=set())

    # -------------------------------------------------------------------------
    def __post_init__(self):
        super().__post_init__()

        self.Name = "EQ-IDX {0:s}".format(self.Symbol)
        self.Country = self.Country or GetVal(self.Exchange, "Country")
        self.Region = self.Region or GetVal(self.Exchange, "Region")

        # --- name of the time-series where historical cash prices are stored
        self.TimeSeries = "IDX-TS {0:s}".format(self.Symbol)

    # -------------------------------------------------------------------------
    @ValueType()
    def Ticker(self, graph):
        return graph(self, "Tickers")[graph("Settings", "Platform")]

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def GetContract(self, graph, del_mth):
        """
        Description:
            Return the EqutyIndexCnt object for a specific delivery month.
        Inputs:
            del_mth - the delivery month in LYY format (as in Z11)
        Returns:
            A string.
        """
        if del_mth in graph(self, "Contracts"):
            symbol = graph(self, "Symbol")
            return EquityIndexCnt.get_name(symbol, del_mth)
        else:
            raise MissingContract("Contract {0:s} not found "
                                  "for {1:s}".format(del_mth, self.Name))

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetCurve(self, graph, start=None, end=None, field=None):
        name = graph(self, "TimeSeries")
        try:
            return TsDbGetCurve(name, start, end, "HLOCV", field)
        except TsNotFound:
            return HlocvCurve() if field is None else Curve()

    # -------------------------------------------------------------------------
    def add_contract(self, cnt_mth, tickers):
        """
        Description:
            Add a new contract or return the existing one.
        Inputs:
            cnt_mth - the contract month in LYY format
            tickers - a dictionary of tickers
        Returns:
            The contract's name.
        """
        if cnt_mth in self.Contracts:
            return GetVal(self, "GetContract", cnt_mth)
        else:
            info = {
                "EquityIndex": self.Name,
                "DeliveryMonth": cnt_mth,
                "Tickers": tickers,
                "RiskProxy": self.Name,
            }
            cnt_obj = EquityIndexCnt(**info)
            try:
                AddObj(cnt_obj)
            except ObjExists:
                obj = GetObj(cnt_obj.Name)
                if obj != cnt_obj:
                    raise RuntimeError("StoredAttrs of existing contract "
                                       "{0:s} don't match those of parent "
                                       "EquityIndex".format(obj.Name))

            self.Contracts.add(cnt_mth)

        return cnt_obj.Name

    # -------------------------------------------------------------------------
    def delete(self):
        sym = self.Symbol
        with ObjDbTransaction("deleting contracts", "SERIALIZABLE"):
            # --- conversion to tuple is needed because the delete method of a
            #     contract removes such contract from the set of contracts of
            #     the asset.
            for cnt_mth in tuple(self.Contracts):
                cnt_name = EquityIndexCnt.get_name.format(sym, cnt_mth)
                try:
                    DelObj(cnt_name)
                except ObjNotFound:
                    print("contract {0:s} not found ?!?".format(cnt_name))


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing

    from ..system import ufo_database
    from ..system import ufo_settings
    from ..system import ufo_holiday_calendar
    from ..system import ufo_currency
    from ..system import ufo_exchange

    ufo_database.prepare_for_test()
    ufo_settings.prepare_for_test()
    ufo_holiday_calendar.prepare_for_test()
    ufo_currency.prepare_for_test()
    ufo_exchange.prepare_for_test()

    sx5e_info = {
        "Symbol": "SX5E",
        "Exchange": "XEUR Exchange",
        "Tickers": {"Bloomberg": "SX5E"},
        "Multiplier": 1.0,
        "ContractSize": 10.0,
    }

    AddIfMissing(EquityIndex(**sx5e_info))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, RDate, DateRange
from onyx.core import Curve, GCurve, Interpolate, CurveUnion
from onyx.core import SetField, DictField, ValueType, Archivable
from onyx.core import MktIndirectionFactory, EnforceArchivableEntitlements

import numpy as np

__all__ = ["Basket"]


###############################################################################
@EnforceArchivableEntitlements("Database", "ArchivedOverwritable")
class Basket(Archivable):
    """
    Container-like object use to represent a basket of equally weighted assets.
    """
    FilterBy = DictField(default={})

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(SetField)
    def Symbols(self, graph):
        pass

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def SymbolsCurve(self, graph, start=None, end=None):
        crv = self.get_history("Symbols", start, end)
        if len(graph(self, "FilterBy")):
            vls = []
            for symbols in crv.values:
                for attr, value in graph(self, "FilterBy").items():
                    symbols = {sym for sym in symbols
                               if graph(sym, attr) == value}
                vls.append(symbols)
            crv = GCurve(crv.dates, vls)
        return crv

    # -------------------------------------------------------------------------
    @ValueType()
    def FilteredSymbols(self, graph):
        symbols = graph(self, "Symbols")
        for attr, value in graph(self, "FilterBy").items():
            symbols = {sym for sym in symbols if graph(sym, attr) == value}
        return symbols

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetCurve(self, graph, start=None, end=None):
        start = start or Date.low_date()
        end = end or graph("Database", "MktDataDate")
        symbols_crv = graph(self, "SymbolsCurve", start, end)

        rts_crv = Curve()
        for date, symbols in symbols_crv:
            new_rts_crv = self.get_returns(graph, date, end, symbols)
            rts_crv = CurveUnion(new_rts_crv, rts_crv)

        returns = rts_crv.values
        returns[0] = 0.0

        # --- the benchkmark is reconstructed from the series of returns
        return Curve(rts_crv.dates, np.cumprod(1.0 + returns))

    # -------------------------------------------------------------------------
    def get_returns(self, graph, start, end, symbols):
        dts = list(DateRange(start + RDate("-1d"), end, "+1d"))
        rts = np.zeros(len(dts) - 1)

        # --- we adjust the start date to make sure that we always have at
        #     least one knot in the curve before start date (this is needed
        #     by Interpolate)
        start_adj = start + RDate("-1w")

        # --- the basket is rewighted daily, hence we just need to calculate
        #     the series of average returns
        for sym in symbols:
            crv = graph(sym, "PricesForRisk").crop(start_adj, end)
            vls = np.array([Interpolate(crv, d) for d in dts])
            rts += vls[1:] / vls[:-1] - 1.0

        num_of_symbols = len(symbols)
        if num_of_symbols:
            rts /= float(num_of_symbols)

        return Curve(dts[1:], rts)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing, Date, RDate, EvalBlock
    from . import ufo_database

    ufo_database.prepare_for_test()

    dt1 = Date.today() + RDate("-2m")
    dt2 = Date.today() + RDate("-1m")

    sym1 = {"aaa", "bbb"}
    sym2 = {"aaa", "ccc"}

    bskt = AddIfMissing(Basket(Name="TestBasket"))

    with EvalBlock() as eb:
        eb.change_value("Database", "ArchivedOverwritable", True)
        bskt.set_dated("Symbols", dt1, sym1)
        bskt.set_dated("Symbols", dt2, sym2)

    return bskt, [dt1, dt2], [sym1, sym2]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, Structure, DateOffset
from onyx.core import AddObj, GetObj, UpdateObj, DelObj, PurgeObj, ObjNotFound
from onyx.core import CreateInMemory, ObjDbTransaction, ObjDbQuery
from onyx.core import ValueType, GetVal, SetVal, UfoBase, ReferenceField
from onyx.core import IntField, FloatField, DateField, SelectField, BoolField
from onyx.core import unique_id

from .tradable_api import AddByInference
from .transaction_api import Transaction, TransactionsBy, TransactionWarning
from .transaction_api import CommitTransaction, DeleteTransaction

from .ufo_forward_cash import ForwardCash
from ..risk.decorators import WithRiskValueTypes

import random
import logging

__all__ = [
    "Trade",
    "TradeError",
    "TradeWarning",
    "CommitTrade",
    "DeleteTrade",
    "ChildrenByBook",
    "TradesBy",
]

logger = logging.getLogger(__name__)

# --- templates for common queries

TRANSACTIONS_STORED = """
SELECT Event, Transaction
FROM (
    SELECT Event, Transaction
    FROM PosEffects
    WHERE Trade=%s AND Book=%s
) tab1 INNER JOIN LATERAL (
    SELECT FROM Objects
    WHERE Name=Transaction AND Data@>'{"Marker": "HEAD"}'::jsonb
    LIMIT 1
) tab2 ON true;
"""

CHILDREN_BY_BOOK = """
SELECT Unit, SUM(Qty) AS tot_qty
FROM PosEffects WHERE Book=%s AND {0:s}<=%s
GROUP BY Unit;"""


###############################################################################
class TradeError(Exception):
    pass


###############################################################################
class TradeWarning(Warning):
    pass


###############################################################################
@WithRiskValueTypes
class Trade(UfoBase):
    """
    Class used to represent trade objects in the system and their effects
    on positions.
    """
    SecurityTraded = ReferenceField(obj_type="TradableObj")
    TradeDate = DateField()
    TradeType = SelectField(default="Buy", options=["Buy", "Sell"])
    Quantity = IntField(default=0, positive=True)
    UnitPrice = FloatField(default=0.0, positive=True)
    PaymentUnit = ReferenceField(obj_type="Currency")
    SettlementDate = DateField()
    Party = ReferenceField(obj_type="Book")
    Counterparty = ReferenceField(obj_type="Book")
    Trader = ReferenceField(obj_type="Trader")
    Broker = ReferenceField(obj_type="Broker")
    BrokerageFee = FloatField(default=0.0, positive=True)
    OtherCosts = FloatField(default=0.0, positive=True)
    Deleted = BoolField(default=False)

    # -------------------------------------------------------------------------
    def __post_init__(self):
        # --- set object's name if not set already
        self.Name = self.Name or Trade.random_name()

        # --- set default values
        self.PaymentUnit = self.PaymentUnit or "USD"

        if self.SettlementDate is None:
            cal = GetVal(self.PaymentUnit, "HolidayCalendar")
            self.SettlementDate = DateOffset(self.TradeDate, "+2b", cal)

        self.Broker = self.Broker or "BROKER-SUSPENSE"

    # -------------------------------------------------------------------------
    @ValueType()
    def Denominated(self, graph):
        """
        A trade is denominated in the PaymentUnit currency.
        """
        return graph(self, "PaymentUnit")

    # -------------------------------------------------------------------------
    @ValueType()
    def TradeInfo(self, graph):
        """
        Return a dictionary {attribute: value}, including all the stored
        attributes.
        """
        attrs = sorted(self._json_fields)
        return dict(zip(attrs, [graph(self, attr) for attr in attrs]))

    # -------------------------------------------------------------------------
    @ValueType()
    def TransactionsCalc(self, graph):
        sec = graph(self, "SecurityTraded")
        trd_date = graph(self, "TradeDate")
        trd_type = graph(self, "TradeType")
        trd_qty = graph(self, "Quantity")
        qty = trd_qty*(1.0 if trd_type == "Buy" else -1.0)
        ccy = graph(self, "PaymentUnit")
        prc = graph(self, "UnitPrice")
        sett_dt = graph(self, "SettlementDate")
        cash = ForwardCash(Currency=ccy, PaymentDate=sett_dt)
        cash = AddByInference(cash, in_memory=True)

        return {
            "BuySell": CreateInMemory(Transaction(
                TransactionDate=trd_date,
                SecurityTraded=sec,
                Quantity=qty,
                Party=graph(self, "Party"),
                Counterparty=graph(self, "Counterparty"),
                Event="BuySell",
                Trade=self.Name
            )),
            "Payment": CreateInMemory(Transaction(
                TransactionDate=trd_date,
                SecurityTraded=cash.Name,
                Quantity=-qty*prc,
                Party=graph(self, "Party"),
                Counterparty=graph(self, "Counterparty"),
                Event="Payment",
                Trade=self.Name
            )),
            "Costs": CreateInMemory(Transaction(
                TransactionDate=trd_date,
                SecurityTraded=cash.Name,
                Quantity=-graph(self, "OtherCosts"),
                Party=graph(self, "Party"),
                Counterparty="TAXNCOSTS",
                Event="Costs",
                Trade=self.Name
            )),
            "Fees": CreateInMemory(Transaction(
                TransactionDate=trd_date,
                SecurityTraded=cash.Name,
                Quantity=-graph(self, "BrokerageFee"),
                Party=graph(self, "Party"),
                Counterparty=graph(self, "Broker"),
                Event="Fees",
                Trade=self.Name
            ))
        }

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        leaves = Structure()
        for transaction in graph(self, "TransactionsCalc").values():
            leaves += graph(transaction, "Leaves")
        return leaves

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm = 0.0
        for leaf, qty in graph(self, "Leaves").items():
            mtm += qty*graph(leaf, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @ValueType()
    def PositionEffects(self, graph):
        pos_effects = []
        for transaction in self.transactions_stored().values():
            pos_effects += graph(transaction, "PositionEffects")
        return pos_effects

    # -------------------------------------------------------------------------
    @ValueType()
    def PositionEffectsCalc(self, graph):
        pos_effects = []
        for transaction in graph(self, "TransactionsCalc").values():
            pos_effects += graph(transaction, "PositionEffects")
        return pos_effects

    # -------------------------------------------------------------------------
    def delete(self):
        if GetVal("Database", "TradesDeletable"):
            transactions = TransactionsBy(trade=self.Name, head_only=False)
            with ObjDbTransaction("Delete Children", "SERIALIZABLE"):
                for transaction in transactions:
                    DelObj(transaction)
        else:
            raise TradeError("Trying to delete a trade without permission")

    # -------------------------------------------------------------------------
    # NB: this needs to be a standard method to make sure that there is never
    #     any caching on trade.Name and trade.Party (for clarity sake we
    #     refrain from making this a property).
    def transactions_stored(self):
        rows = ObjDbQuery(TRANSACTIONS_STORED,
                          (self.Name, self.Party), attr="fetchall")

        stored = {}
        for row in rows:
            try:
                stored[row.event].append(row.transaction)
            except AttributeError:
                stored[row.event] = [stored[row.event], row.transaction]
            except KeyError:
                stored[row.event] = row.transaction

        return stored

    # -------------------------------------------------------------------------
    @classmethod
    def random_name(cls):
        return "TmpTrd-{0:>07d}".format(random.randrange(0, 10000000, 1))

    # -------------------------------------------------------------------------
    @classmethod
    def format_name(cls, date, trd_id):
        return "TRD {0:s} {1:s}".format(date.strftime("%Y%m%d"), trd_id)


# -----------------------------------------------------------------------------
def PreTradeCompliance(trade):
    """
    Description:
        This function is ment to raise an exception for any violation
        of pre-trade compliance checks.
    Inpots:
        trade - the trade to be checked
    Returns:
        None.
    """
    # --- price validation
    GetVal(trade, "MktValUSD")

    # --- basic risk validation
    GetVal(trade, "Deltas")


# -------------------------------------------------------------------------
def CommitTrade(trade, name=None):
    """
    Description:
        Create/amend a trade by creting/amending the associated transactions.
    Inputs:
        trade - the instance of Trade to commit
        name  - an optional trade name
    Returns:
        A Trade instance.
    """
    # --- clone security traded: AddByInference is needed to generate the
    #     proper ImpliedName.
    sec = AddByInference(GetObj(trade.SecurityTraded).clone())

    # --- get trade info before reloading the trade
    info = GetVal(trade, "TradeInfo")
    info["SecurityTraded"] = sec.Name

    try:
        # --- reload security: this is needed to avoid basing decisions on
        #     attribute values that might have been set in memory
        trade = GetObj(trade.Name, refresh=True)

    except ObjNotFound:
        # --- trade not found, proceed as new trade
        trade.SecurityTraded = sec.Name

        # --- create trade and overwrite a few stored attributes
        name = name or trade.format_name(Date.today(), unique_id(8))
        trade = CreateInMemory(trade.clone(Name=name))

        PreTradeCompliance(trade)

        logger.info("creating new trade {0:s}".format(trade.Name))

        calculated = GetVal(trade, "TransactionsCalc")
        with ObjDbTransaction("Trade Insert", level="SERIALIZABLE"):
            for transaction in calculated.values():
                CommitTransaction(transaction)
            AddObj(trade)

    else:
        # --- need to amend?
        same = True
        for attr, value in info.items():
            same = (value == getattr(trade, attr))
            if not same:
                break
        if same:
            raise TradeWarning("{0:s}, TradeCommit found nothing to "
                               "amend on existing trade.".format(trade.Name))

        logger.info("amending trade {0:s}".format(trade.Name))

        # --- we fetch stored transactions before overwriting the stored
        #     attributes of the trade (this is essential since amending the
        #     Party attribute would prevent fetching the correct stored
        #     transactions).
        stored = trade.transactions_stored()  # this queries the backend

        if not len(stored):
            raise TradeError("{0:s}, error loading "
                             "transactions stored.".format(trade.Name))

        # --- overwrite trade stored attributes with new ones
        for attr, value in info.items():
            SetVal(trade, attr, value)

        PreTradeCompliance(trade)

        calculated = GetVal(trade, "TransactionsCalc")
        with ObjDbTransaction("Trade Amend", level="SERIALIZABLE"):
            UpdateObj(trade.Name)
            for event, transaction in calculated.items():
                # --- overwrite the name of the calculated transaction with
                #     that of the stored one
                try:
                    stored_name = stored[event]
                except KeyError:
                    raise TradeError("{0:s}, trying to amend a trade with "
                                     "an aged transaction.".format(trade.Name))
                try:
                    CommitTransaction(transaction, stored_name)
                except TransactionWarning as warning:
                    # --- ignore warnings
                    logger.info(warning)

    return trade


# -------------------------------------------------------------------------
def DeleteTrade(trade_name):
    """
    Description:
        Delete a trade, by deleting all associated transactions and marking it
        as deleted.
    Inputs:
        trade_name - the name of the trade to delete
    Returns:
        None.
    """
    # --- reload security: this is needed to avoid basing decisions on
    #     attribute values that might have been set in memory
    trade = GetObj(trade_name, refresh=True)

    with ObjDbTransaction("Trade Delete", level="SERIALIZABLE"):
        if trade.TimeCreated >= Date.today():
            # --- same day delete: get rid of the trade and of all associated
            #     transactions
            PurgeObj(trade)
        else:
            trade.Deleted = True
            UpdateObj(trade)
            for transaction in trade.transactions_stored().values():
                if isinstance(transaction, list):
                    for sub_transaction in transaction:
                        DeleteTransaction(sub_transaction)
                else:
                    DeleteTransaction(transaction)


# -----------------------------------------------------------------------------
def ChildrenByBook(book, pos_date, by_time_created=False):
    """
    Description:
        Return all children of a given book as of a given date.
    Inputs:
        book            - the book
        pos_date        - the positions date
        by_time_created - if True, amendments are reflected on the date they
                          were booked, not on the date of the original trade.
    Returns:
        A Structure.
    """
    if by_time_created:
        query = CHILDREN_BY_BOOK.format("TimeCreated")
    else:
        query = CHILDREN_BY_BOOK.format("TradeDate")

    rows = ObjDbQuery(query, parms=(book, pos_date.eod()), attr="fetchall")
    children = Structure()
    for row in rows:
        # --- exclude children with zero quantity
        if row.tot_qty != 0.0:
            children[row.unit] = row.tot_qty

    return children


# -----------------------------------------------------------------------------
def TradesBy(book=None, trader=None,
             security=None, sectype=None, date_range=None):
    """
    Description:
        Return all trades matching the given criteria.
    Inputs:
        book       - a given book
        trader     - a given trader
        security   - a given security
        sectype    - a given security type
        date_range - a tuple of start and end date
    Returns:
        A list of trade names.
    """
    extquery = []
    criteria = []

    if book is not None:
        extquery.append("Book=%s")
        criteria.append(book)

    if security is not None:
        extquery.append("Unit=%s")
        criteria.append(security)

    if sectype is not None:
        extquery.append("UnitType=%s")
        criteria.append(sectype)

    if date_range is not None:
        extquery.append("TradeDate BETWEEN %s AND %s")
        criteria.append(date_range[0])
        criteria.append(date_range[1])

    if len(criteria):
        query = ("SELECT DISTINCT(Trade) "
                 "FROM PosEffects WHERE {0:s};").format(" AND ".join(extquery))
        rows = ObjDbQuery(query, parms=criteria, attr="fetchall")

    else:
        query = "SELECT DISTINCT(Trade) FROM PosEffects;"
        rows = ObjDbQuery(query, attr="fetchall")

    # --- postprocessing filters
    if trader is not None:
        # --- Trader is not part of PosEffects table: run filter as a
        #     post-process
        def post_proc(trade_name):
            trade = GetObj(trade_name)
            return trade.Trader == trader and not trade.Deleted
    else:
        def post_proc(trade_name):
            trade = GetObj(trade_name)
            return not trade.Deleted

    return [row.trade for row in rows if post_proc(row.trade)]


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import Date, AddIfMissing

    from . import ufo_book
    from . import ufo_trader
    from . import ufo_broker
    from . import ufo_forward_cash

    from ..equity import ufo_equity_cash
    from ..equity import ufo_equity_cfd

    books = ufo_book.prepare_for_test()
    traders = ufo_trader.prepare_for_test()
    brokers = ufo_broker.prepare_for_test()
    eqcash = ufo_equity_cash.prepare_for_test()
    eqcfds = ufo_equity_cfd.prepare_for_test()

    ufo_forward_cash.prepare_for_test()

    trades = [
        # --- this is a cash trade
        {
            "SecurityTraded": eqcash[0],
            "TradeDate": Date.today(),
            "TradeType": "Sell",
            "Quantity": 1000,
            "Party": books[0],
            "Counterparty": books[2],
            "Trader": traders[0],
            "Broker": brokers[0],
            "UnitPrice": 9.1*1.5/1.15,
            "PaymentUnit": "EUR",
        },
        # --- this is a CFD trade
        {
            "SecurityTraded": eqcfds[0],
            "TradeDate": Date.today(),
            "TradeType": "Sell",
            "Quantity": 1000,
            "Party": books[0],
            "Counterparty": books[2],
            "Trader": traders[0],
            "Broker": brokers[0],
        }
    ]
    trades = [AddIfMissing(Trade(**trade_info), True) for trade_info in trades]

    return [trade.Name for trade in trades]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Structure, DateRange
from onyx.core import GraphNodeVt, ReferenceField, DateField, StringField

from .tradable_api import AgingTradableObj, AddByInference, HashStoredAttrs
from .ufo_forward_cash import ForwardCash


###############################################################################
class AmortizedCash(AgingTradableObj):
    """
    Tradable class that represents the depreciation/amortization of an asset or
    liability.
    For instance, to book an amortized fee one will have to book a buy trade on
    AmortizedCash with UnitPrice=1.0 and Quantity=FeeAmount. The cash-flow
    takes place on the trade's settlement date but the fee hits the P&L on the
    basis of the amortization schedule.
    """
    Currency = ReferenceField(obj_type="Currency")
    StartDate = DateField()
    EndDate = DateField()
    DateRule = StringField(default="+1b")

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def Dates(self, graph):
        start = graph(self, "StartDate")
        end = graph(self, "EndDate")
        rule = graph(self, "DateRule")
        return list(DateRange(start, end, rule))

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def AmortizedAmount(self, graph):
        return 1.0 / float(len(graph(self, "Dates")))

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def Leaves(self, graph):
        pos_date = graph("Database", "PositionsDate")
        ccy = graph(self, "Currency")
        dts = graph(self, "Dates")
        amt = graph(self, "AmortizedAmount")
        qty = 1.0 - amt*sum([1.0 for d in dts if d <= pos_date])
        fwd_cash = ForwardCash(Currency=ccy, PaymentDate=pos_date)
        fwd_cash = AddByInference(fwd_cash, in_memory=True)
        return Structure({fwd_cash.Name: qty})

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def MktValUSD(self, graph):
        mtm = 0.0
        for sec, qty in graph(self, "Leaves").items():
            mtm += qty*graph(sec, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def MktVal(self, graph):
        cross = "{0:3s}/USD".format(graph(self, "Currency"))
        return graph(self, "MktValUSD") / graph(cross, "Spot")

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def ExpirationDate(self, graph):
        return graph(self, "EndDate")

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def NextTransactionDate(self, graph):
        return graph(self, "EndDate")

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def TradeTypes(self, graph):
        mapping = super().TradeTypes
        mapping.update({
            "Expiry": "ExpirySecurities",
        })
        return mapping

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def ExpectedTransaction(self, graph):
        return "Expiry"

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def ExpirySecurities(self, graph):
        return []

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        start = self.StartDate.strftime("%d%m%y")
        end = self.EndDate.strftime("%d%m%y")
        mush = HashStoredAttrs(self, 4)
        return ("AMT {0:3s} {1:6s} {2:6s} {3:4s} "
                "{{0:2d}}").format(self.Currency, start, end, mush)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Curve, Knot, UfoBase
from onyx.core import ReferenceField, StringField, ValueType, RetainedFactory
from onyx.core import TsDbGetRowBy, TsDbGetCurve, TsDbUpsertCurve, TsNotFound

__all__ = ["CurrencyCross"]


###############################################################################
class CurrencyCross(UfoBase):
    """
    Class used to represent a cross between two currencies (i.e. FX rates).
    """
    # --- conversion is Currency1 -> Currency2
    Currency1 = ReferenceField(obj_type="Currency")
    Currency2 = ReferenceField(obj_type="Currency")
    FxTimeSeries = StringField()

    # -------------------------------------------------------------------------
    def __post_init__(self):
        self.Name = "{0:3s}/{1:3s}".format(self.Currency1, self.Currency2)
        self.FxTimeSeries = "FX-TS {0:s}".format(self.Name)

    # -------------------------------------------------------------------------
    @ValueType()
    def Ticker(self, graph):
        platform = graph("Settings", "Platform")
        if platform == "Bloomberg":
            return "{0:3s}{1:3s} CURNCY".format(*self.Name.split("/"))
        else:
            raise NotImplementedError("Unrecognized "
                                      "platform {0:s}".format(platform))

    # -------------------------------------------------------------------------
    @RetainedFactory()
    def Spot(self, graph):
        """
        Return the official close value as of MktDataDate (or the most recent
        close if ForceStrict is False).
        """
        day = graph("Database", "MktDataDate")
        name = graph(self, "FxTimeSeries")
        row = TsDbGetRowBy("Curves", name,
                           day, graph("Database", "ForceStrict"))
        return row[2]

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def SpotByDate(self, graph, day):
        """
        Return the official close value as of the specified date (or the most
        recent close if ForceStrict is False).
        """
        name = graph(self, "FxTimeSeries")
        row = TsDbGetRowBy("Curves", name,
                           day, graph("Database", "ForceStrict"))
        return row[2]

    # -------------------------------------------------------------------------
    @ValueType()
    def Last(self, graph):
        """
        Return the knot with the most recent close value irrespective of
        MktDataDate.
        """
        row = TsDbGetRowBy("Curves", graph(self, "FxTimeSeries"))
        return Knot(row.date, row.value)

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetCurve(self, graph, start=None, end=None):
        try:
            return TsDbGetCurve(graph(self, "FxTimeSeries"), start, end)
        except TsNotFound:
            return Curve()

    # -------------------------------------------------------------------------
    def update_timeseries(self, data):
        if isinstance(data, (tuple, list)):
            data = Curve([data[0]], [data[1]])

        TsDbUpsertCurve(self.FxTimeSeries, data)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import Date, GetObj, AddIfMissing
    from . import ufo_database
    from . import ufo_settings
    from . import ufo_currency

    ufo_database.prepare_for_test()
    ufo_settings.prepare_for_test()
    ufo_currency.prepare_for_test()

    AddIfMissing(CurrencyCross(Currency1="USD", Currency2="USD"))
    AddIfMissing(CurrencyCross(Currency1="EUR", Currency2="USD"))
    AddIfMissing(CurrencyCross(Currency1="GBP", Currency2="USD"))

    marks = {
        "USD/USD": 1.00,
        "EUR/USD": 1.15,
        "GBP/USD": 1.50,
    }

    for cross, value in marks.items():
        cross = GetObj(cross)
        cross.update_timeseries([Date.today(), value])

    return marks



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import RDate, Curve, DateOffset, CurveIntersect, CurveUnion
from onyx.core import ObjNotFound, GetVal

import numpy as np

__all__ = ["get_curve_usd", "prices_for_risk"]


# -----------------------------------------------------------------------------
def get_curve_usd(name, start, end):
    """
    Description:
        Given an asset, return the curve with close prices converted to USD.
    Inputs:
        name  - the asset's name
        start - the start date
        end   - the end date
    Returns:
        A Curve.
    """
    # --- we invoke GetVal with start and end set to None and crop the returned
    #     curve so that curve caching remains effective when start and end
    #     dates are changed (at the cost of a larger memory footprint).
    prc = GetVal(name, "GetCurve", field="Close").crop(start, end)
    if len(prc) == 0:
        return Curve()

    mul = GetVal(name, "Multiplier")
    cross = "{0:3s}/USD".format(GetVal(name, "Denominated"))
    fx = GetVal(cross, "GetCurve").crop(start, end)
    prc, fx = CurveIntersect([prc, fx])
    return mul*prc*fx


# -----------------------------------------------------------------------------
def merge_pirces(crv1, crv2):
    """
    Description:
        Merge two timeseries, from their log-return representation
    Inputs:
        crv1 - first curve, takes precedence
        crv2 - second curve, fills missing data in crv1
    Returns:
        A Curve.
    """
    # --- log-returns
    ret1 = Curve(crv1.dates[1:], np.diff(np.log(crv1.values)))
    ret2 = Curve(crv2.dates[1:], np.diff(np.log(crv2.values)))

    # --- combine returns, with priority to those from ret1
    ret = CurveUnion(ret1, ret2)

    # --- reconstruct series of prices from series of returns
    last = crv1.back.value
    rec = np.roll(last*np.exp(np.cumsum(-ret.values[::-1]))[::-1], -1)
    rec[-1] = last

    return Curve(ret.dates, rec)


# -----------------------------------------------------------------------------
def prices_for_risk(name, start, end, strict=True):
    """
    Description:
        Return close prices converted to USD, sourcing them from a proxy
        security when the time series is shorter than needed.
    Inputs:
        name  - the asset's name
        start - start date
        end   - end date
    Returns:
        A Curve.
    """
    # --- make sure that start and end dates are business days for the given
    #     security
    try:
        cal = GetVal(name, "HolidayCalendar")
    except ObjNotFound:
        cal = "USD Calendar"
    start = DateOffset(start, "+0b", cal)
    end = DateOffset(end, "-0b", cal)

    # --- fetch a few extra days (1 week), just to make sure we have enough
    #     data-points
    crv_start = start + RDate("-1w")
    prcs = get_curve_usd(name, crv_start, end)

    if strict and len(prcs) and prcs.back.date != end:
        raise RuntimeError("Missing prices at the end of the curve "
                           "for {0:s}: {1!s}".format(name, prcs.back.date))

    if len(prcs) and prcs.front.date <= start:
        return prcs.crop(start, end)

    proxy = GetVal(name, "RiskProxy")

    if proxy is None:
        raise RuntimeError("We need a proxy for {0:s} to be able "
                           "to reconstruct the time series for the "
                           "range {1!s} - {2!s}".format(name, start, end))
    else:
        prcs_proxy = prices_for_risk(proxy, crv_start, end, strict)

        if len(prcs):
            prcs = merge_pirces(prcs, prcs_proxy)
        else:
            prcs = prcs_proxy

        if prcs.front.date <= start:
            return prcs.crop(start, end)

    # --- we should never end up here: if proxy is not sufficient, an exception
    #     should be raised by prices_for_risk(proxy, start_crv, end, strict)
    raise RuntimeError("Something didn't work properly with "
                       "{0:s} and proxy {1:s}".format(name, proxy))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################
"""
This module defines the basic API and abstract base classes for all tradable
objects (i.e. securities we can have a position on).
"""

from onyx.core import Structure, GetObj, AddObj, ObjNotFound
from onyx.core import UfoBase, CreateInMemory, ValueType, b36encode

import abc
import re
import hashlib

__all__ = [
    "TradableError",
    "NamedByInference",
    "TradableObj",
    "AgingTradableObj",
    "HashStoredAttrs",
    "InferredName",
    "AddByInference",
]


###############################################################################
class TradableError(Exception):
    pass


###############################################################################
class NamedByInference(UfoBase):
    """
    Abstract base class for all objects that require automatic naming. These
    are generally children of tradables but never tradables themselves.
    """
    # -------------------------------------------------------------------------
    def __post_init__(self):
        self.Name = None

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def ImpliedName(self):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        return Structure({self.Name: 1.0})


###############################################################################
class TradableObj(NamedByInference):
    """
    Abstract base class for all tradable objects (securities). This provides
    the basic interface that each tradable needs to expose.
    """
    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def MktVal(self, graph):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def MktValUSD(self, graph):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @ValueType()
    def TradeTypes(self, graph):
        return {
            "Buy": "BuySecurities",
            "Sell": "SellSecurities",
        }

    # -------------------------------------------------------------------------
    @ValueType()
    def BuySecurities(self, graph):
        return [{"Security": self, "Quantity": 1.0}]

    # -------------------------------------------------------------------------
    @ValueType()
    def SellSecurities(self, graph):
        return [{"Security": self, "Quantity": -1.0}]

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def ExpectedSecurities(self, graph, key):
        """
        Return a list of securities and quantities as follows:
        [
            {"Security": sec_1, "Quantity": qty_1},
            {"Security": sec_2, "Quantity": qty_2},
            ...
        ]
        """
        return graph(self, graph(self, "TradeTypes")[key])


###############################################################################
class AgingTradableObj(TradableObj):
    """
    Abstract base class for all tradable objects (securities) that require an
    aging mechanism.

    The aging mechanism must be implemented as follows:
        - ExpectedSecurities returns a list of dictionaries with expected
          security and expected quantity:
              [{"Security": sec, "Quantity": qty}, ...]
          The list is generated based on the value of ExpectedTransaction and
          the TradeTypes mapping.
        - ExpectedTransaction returns a string with the transaction expected to
          take place at the NextTransactionDate (such as Exercise, Expire,
          Pay, etc). Such string must be defined in TradeTypes.
        - NextTransactionDate returns the future date when a new event takes
          place.
    """
    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def NextTransactionDate(self, graph):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def ExpectedTransaction(self, graph):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def ExpectedSecurities(self, graph, key=None):
        if key is None:
            key = graph(self, "ExpectedTransaction")
        return graph(self, graph(self, "TradeTypes")[key])


# -----------------------------------------------------------------------------
def HashStoredAttrs(obj, nchars, skip=None):
    """
    Description:
        Return a base36 representation of the stored attributes of an object.
    Inputs:
        obj    - instance of a class derived from UfoBase
        nchars - number of characters in the hash
        skip   - a set of attributes to skip
    Returns:
        A string.
    """
    skip = skip or {}
    attributes = set(obj._json_fields).difference(skip)

    md5_hash = hashlib.md5()
    for attr in sorted(attributes):
        md5_hash.update(bytes("{0!r}".format(getattr(obj, attr)), "utf-8"))

    mush = b36encode(md5_hash.hexdigest())
    nchars = min(nchars, len(mush))

    return mush[:nchars]


# -----------------------------------------------------------------------------
def InferredName(obj):
    """
    Description:
        Generate a unique inferred name for a security using its ImpliedName
        method.
    Inputs:
        obj - instance of a class derived from UfoBase
    Returns:
        A string.
    """
    if not isinstance(obj, NamedByInference):
        raise TradableError("InferredName error for {0!s}.\nOnly"
                            "objects derived from NamedByInference "
                            "can have an InferredName".format(type(obj)))

    imp_name = obj.ImpliedName

    # --- check if implied name supports collision bit
    match = re.search("{0:[1-9]d}", imp_name)
    if match:
        for k in range(10**int(match.group()[3:-2])):
            name = imp_name.format(k)
            # --- here we check if an object with this name already exists in
            #     database and if such object is the same as obj
            try:
                if obj == GetObj(name):
                    return name
                else:
                    continue
            except ObjNotFound:
                return name
    else:
        try:
            if obj == GetObj(imp_name):
                return imp_name
        except ObjNotFound:
            return imp_name

    # --- nothing worked, raise an exception
    raise TradableError("ImpliedName collision problem for "
                        "<{0:s} - {1:s}>".format(imp_name, obj.ObjType))


# -----------------------------------------------------------------------------
def AddByInference(obj, in_memory=False):
    """
    Description:
        Add a security based on it's InferredName or return the current
        instance if the security exists already.
    Inputs:
        obj       - instance of a class derived from UfoBase
        in_memory - if True, create object in memory
    Returns:
        The object's instance.
    """
    obj.Name = InferredName(obj)

    if in_memory:
        return CreateInMemory(obj)
    else:
        try:
            return GetObj(obj.Name, refresh=True)
        except ObjNotFound:
            return AddObj(obj)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetObj, GetVal, ObjDbTransaction
from .transaction_api import TransactionsBy, AgeTransaction

import logging

__all__ = ["age_portfolio"]

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
def age_portfolio(port):
    """
    Description:
        Age postions for a given portfolio as of Database->PositionsDate.
    Inputs:
        port - the portfolio whose positions we want to age
    """
    books = GetVal(port, "Books")
    kids = set.union(*[set(GetVal(book, "Children").keys()) for book in books])
    pos_date = GetVal("Database", "PositionsDate")

    maybe_more_to_do = False

    for kid in kids:
        try:
            ntd = GetVal(kid, "NextTransactionDate")
        except AttributeError:
            continue

        if ntd == pos_date:
            # --- lookup affected transactions
            transactions = set()
            for book in books:
                trs_by_book = TransactionsBy(security=kid, book=book)
                transactions.update(set(trs_by_book))

            if len(transactions):
                event = GetVal(kid, "ExpectedTransaction")
                secs = GetVal(kid, "ExpectedSecurities", event)
                logger.debug(kid, event, secs)
                logger.debug(sorted(transactions))

                tag = "Aging of {0:s}".format(kid)
                with ObjDbTransaction(tag, level="SERIALIZABLE"):
                    for transaction in sorted(transactions):
                        AgeTransaction(GetObj(transaction), pos_date)

                maybe_more_to_do = True

    return maybe_more_to_do



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Curve, CurveUnion, Date2LYY, DateOffset, DateRange
from onyx.core import ValueType, GetVal
from onyx.core import IntField, DateField, SelectField, ReferenceField

from ..system.tradable_api import NamedByInference, HashStoredAttrs

import collections

__all__ = ["CommodNrby"]

AVERAGING_TYPES = ["DAILYCALDAYS", "DAILYBIZDAYS", "LAST"]


###############################################################################
class CommodNrby(NamedByInference):
    """
    CommodNrby class, used to provide nearby information to COMMOD tradable
    objects.
    """
    Asset = ReferenceField(obj_type="CommodAsset")
    StartDate = DateField()
    EndDate = DateField()
    AvgType = SelectField(options=AVERAGING_TYPES)
    RollType = IntField(default=0)

    # -------------------------------------------------------------------------
    @ValueType()
    def StartDateRule(self, graph):
        avg_type = graph(self, "AvgType")
        if avg_type == "DAILYCALDAYS":
            return None
        elif avg_type == "DAILYBIZDAYS":
            return "+0b"
        elif avg_type == "LAST":
            asset = graph(self, "Asset")
            return "+0J{0:s}".format(graph(asset, "SettDateRule"))

    # -------------------------------------------------------------------------
    @ValueType()
    def EndDateRule(self, graph):
        avg_type = graph(self, "AvgType")
        if avg_type == "DAILYCALDAYS":
            return None
        elif avg_type == "DAILYBIZDAYS":
            return "-0b"
        elif avg_type == "LAST":
            asset = graph(self, "Asset")
            return "+0J{0:s}".format(graph(asset, "SettDateRule"))

    # -------------------------------------------------------------------------
    @ValueType()
    def IncrementRule(self, graph):
        avg_type = graph(self, "AvgType")
        if avg_type == "DAILYCALDAYS":
            return "+1d"
        elif avg_type == "DAILYBIZDAYS":
            return "+1b"
        elif avg_type == "LAST":
            asset = graph(self, "Asset")
            offset = graph(asset, "NrbyOffset")
            return "+{0:d}m+0J{1:s}".format(1 + offset,
                                            graph(asset, "SettDateRule"))

    # -------------------------------------------------------------------------
    @ValueType()
    def NrbyStartDate(self, graph):
        return DateOffset(graph(self, "StartDate"),
                          graph(self, "StartDateRule"),
                          graph(graph(self, "Asset"), "HolidayCalendar"))

    # -------------------------------------------------------------------------
    @ValueType()
    def NrbyEndDate(self, graph):
        return DateOffset(graph(self, "EndDate"),
                          graph(self, "EndDateRule"),
                          graph(graph(self, "Asset"), "HolidayCalendar"))

    # -------------------------------------------------------------------------
    @ValueType()
    def NearbyDateRule(self, graph):
        """
        Date rule used to determine the relevant nearby contract given a date.
        """
        return "+{0:d}m+0J".format(graph(self, "RollType") +
                                   graph(graph(self, "Asset"), "NrbyOffset"))

    # -------------------------------------------------------------------------
    @ValueType()
    def DataCurve(self, graph):
        """
        The data curve is created by merging two different pieces:
         1) the setteled part of the curve
         2) the open part of the curve
        """
        pd = graph("Database", "PricingDate")
        sd = graph(self, "NrbyStartDate")
        ed = graph(self, "NrbyEndDate")

        asset = graph(self, "Asset")
        cal = graph(asset, "HolidayCalendar")

        sett_rule = graph(asset, "SettDateRule")
        inc_rule = graph(self, "IncrementRule")
        nrby_rule = graph(self, "NearbyDateRule")

        crv = Curve()

        if sd <= pd:
            # --- this is the part of curve that has settled
            cnts = collections.defaultdict(list)
            for d in DateRange(sd, min(pd, ed), inc_rule, cal):
                # --- find the nth (as defined by RollType) nearby contract as
                #     of a given date
                cnt = DateOffset(d, nrby_rule, cal)
                if d > DateOffset(cnt, sett_rule, cal):
                    cnt = DateOffset(cnt, "+1m", cal)

                cnt = graph(asset, "GetContract", Date2LYY(cnt))
                cnts[cnt].append(d)

            # --- build-up the curve
            for cnt, dts in cnts.items():
                cnt_crv = graph(cnt, "GetCurve",
                                dts[0], dts[-1], field="Close")
                crv = CurveUnion(crv, cnt_crv)

            # --- move forward start date
            if len(crv):
                sd = DateOffset(crv.back.date, inc_rule, cal)
            else:
                sd = pd

        # --- this is the open part of the curve
        for d in DateRange(sd, ed, inc_rule, cal):
            # --- find the nth (as defined by RollType) nearby contract as of
            #     a given date
            cnt = DateOffset(d, nrby_rule, cal)
            if d > DateOffset(cnt, sett_rule, cal):
                cnt = DateOffset(cnt, "+1m", cal)

            crv[d] = graph(graph(asset, "GetContract", Date2LYY(cnt)), "Spot")

        return crv

    # -------------------------------------------------------------------------
    @ValueType()
    def AverageValue(self, graph):
        crv = graph(self, "DataCurve")
        if len(crv):
            return crv.values.mean()
        else:
            return 0.0

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        mkt = GetVal(self.Asset, "Market")
        sym = GetVal(self.Asset, "Symbol")
        mush = HashStoredAttrs(self, 8)
        return "CmdNRBY {0:s} {1:s} {2:8s} {{0:2d}}".format(mkt, sym, mush)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, RDate, UfoBase, ValueType, ReferenceField

__all__ = ["Risk"]


###############################################################################
class Risk(UfoBase):
    """
    singleton class used to capture risk parameters.
    """
    RefIndex = ReferenceField(obj_type="EquityIndex")

    # -------------------------------------------------------------------------
    @ValueType()
    def StartDate(self, graph):
        return Date(2008, 6, 1)

    # -------------------------------------------------------------------------
    @ValueType()
    def EndDate(self, graph):
        return graph("Database", "PricingDate") + RDate("-1b")

    # -------------------------------------------------------------------------
    @ValueType()
    def VaRStartDate(self, graph):
        return graph(self, "VaREndDate") + RDate("-5y-0b")

    # -------------------------------------------------------------------------
    @ValueType()
    def VaREndDate(self, graph):
        return graph(self, "EndDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def BetaStartDate(self, graph):
        return graph(self, "BetaEndDate") + RDate("-1y-0b")

    # -------------------------------------------------------------------------
    @ValueType()
    def BetaEndDate(self, graph):
        return graph(self, "EndDate")


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import GetVal, AddIfMissing
    from ..system.ufo_shadow_book import ShadowBook
    from ..system import ufo_database
    from ..system import trade_api

    AddIfMissing(Risk(Name="Risk"))

    ufo_database.prepare_for_test()
    trades = trade_api.prepare_for_test()

    book = AddIfMissing(ShadowBook(Name="TestBook", Denominated="GBP"))

    for trade in trades:
        book.Children += GetVal(trade, "Leaves")

    return [book]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Structure, UfoBase, ValueType
from onyx.core import ReferenceField, StringField, SelectField

from .trade_api import ChildrenByBook
from ..risk.decorators import WithRiskValueTypes

__all__ = ["Book"]


###############################################################################
@WithRiskValueTypes
class Book(UfoBase):
    """
    Class used to represent a trading book (either internal or external).
    """
    Group = ReferenceField(obj_type="Group")
    Denominated = ReferenceField(obj_type="Currency")
    BookType = SelectField(options=["ProfitCenter", "Counterparty"])
    DisplayName = StringField()
    Account = StringField()

    # -------------------------------------------------------------------------
    @ValueType()
    def Books(self, graph):
        return {self.Name}

    # -------------------------------------------------------------------------
    @ValueType()
    def Children(self, graph):
        pos_dt = graph("Database", "PositionsDate")
        return ChildrenByBook(graph(self, "Name"), pos_dt)

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        leaves = Structure()
        for child, qty in graph(self, "Children").items():
            leaves += qty*graph(child, "Leaves")
        return leaves

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm = 0.0
        for leaf, qty in graph(self, "Leaves").items():
            mtm += qty*graph(leaf, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        denominated = graph(self, "Denominated")
        spot_fx = graph("{0:3s}/USD".format(denominated), "Spot")
        return graph(self, "MktValUSD") / spot_fx


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing
    from . import ufo_currency
    from . import ufo_currency_cross
    from . import ufo_group

    ufo_currency.prepare_for_test()
    ufo_currency_cross.prepare_for_test()
    groups = ufo_group.prepare_for_test()

    books = [
        AddIfMissing(Book(Name="TEST_BOOK1", Group=groups[0],
                          Denominated="USD", BookType="ProfitCenter")),
        AddIfMissing(Book(Name="TEST_BOOK2", Group=groups[0],
                          Denominated="USD", BookType="ProfitCenter")),
        AddIfMissing(Book(Name="TEST_CPTY1",
                          Denominated="USD", BookType="Counterparty")),
        AddIfMissing(Book(Name="TAXNCOSTS",
                          Denominated="USD", BookType="Counterparty")),
    ]

    return [book.Name for book in books]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import ValueType, UfoBase, ObjDbQuery
from onyx.core import ReferenceField, StringField

from ..risk.decorators import WithRiskValueTypes

import json

__all__ = ["Group"]

GET_ITEMS = "SELECT Name FROM Objects WHERE ObjType=Trader AND Data @> %s;"


###############################################################################
@WithRiskValueTypes
class Group(UfoBase):
    """
    Class used to represent a trading group (here defined as group of traders).
    """
    Portfolio = ReferenceField(obj_type="Portfolio")
    LongName = StringField()

    # -------------------------------------------------------------------------
    @ValueType()
    def Denominated(self, graph):
        return graph(graph(self, "Portfolio"), "Denominated")

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        return graph(graph(self, "Portfolio"), "Leaves")

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        return graph(graph(self, "Portfolio"), "MktValUSD")

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        return graph(graph(self, "Portfolio"), "MktVal")

    # -------------------------------------------------------------------------
    @ValueType()
    def Traders(self, graph):
        parms = (json.dumps({"Group": graph(self, "Name")}), )
        results = ObjDbQuery(GET_ITEMS, parms, attr="fetchall")
        return sorted([res.name for res in results])


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing
    from . import ufo_portfolio

    portfolios = ufo_portfolio.prepare_for_test()

    groups = [
        AddIfMissing(Group(Name="TEST_GROUP", Portfolio=portfolios[0]))
    ]

    return [group.Name for group in groups]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetVal, GetObj, Structure, ChildrenSet, OnyxTestCase

from onyx.system.ufo_cash_balance import CashBalance
from onyx.system import ufo_forward_cash

import unittest


###############################################################################
class UnitTest(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.fwdusd, self.fwdeur = ufo_forward_cash.prepare_for_test()[:2]

    # -------------------------------------------------------------------------
    def test_Leaves(self):
        self.assertEqual(GetVal(self.fwdusd, "Leaves"),
                         Structure([(self.fwdusd, 1.0)]))

    # -------------------------------------------------------------------------
    def test_MktVal(self):
        self.assertEqual(GetVal(self.fwdusd, "MktVal"), 1.0)
        self.assertEqual(GetVal(self.fwdeur, "MktVal"), 1.0)

    # -------------------------------------------------------------------------
    def test_MktValUSD(self):
        self.assertEqual(GetVal(self.fwdusd, "MktValUSD"), 1.00)
        self.assertEqual(GetVal(self.fwdeur, "MktValUSD"), 1.15)

    # -------------------------------------------------------------------------
    def test_TradeTypes(self):
        tt = {
            "Pay/Receive": "PayReceiveSecurities",
            "Buy": "BuySecurities",
            "Sell": "SellSecurities"
        }
        self.assertEqual(GetVal(self.fwdusd, "TradeTypes"), tt)

    # -------------------------------------------------------------------------
    def test_ExpectedSecurities(self):
        sec_fwdusd = GetObj(self.fwdusd)
        self.assertEqual(GetVal(self.fwdusd, "ExpectedSecurities", "Buy"),
                         [{"Security": sec_fwdusd, "Quantity": 1.0}])
        self.assertEqual(GetVal(self.fwdusd, "ExpectedSecurities", "Sell"),
                         [{"Security": sec_fwdusd, "Quantity": -1.0}])
        self.assertEqual(GetVal(self.fwdusd,
                                "ExpectedSecurities", "Pay/Receive"),
                         [{"Security": CashBalance(Currency="USD"),
                           "Quantity": 1.0}])

    # -------------------------------------------------------------------------
    def test_Children(self):
        kids = ChildrenSet((self.fwdusd, "MktValUSD"), "Spot")
        self.assertEqual(kids, {"USD/USD"})
        kids = ChildrenSet((self.fwdeur, "MktValUSD"), "Spot")
        self.assertEqual(kids, {"EUR/USD"})


if __name__ == "__main__":
    from onyx.core import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Interpolate, CalcTerm
from onyx.core import ReferenceField, CurveField, ValueType, Archivable
from onyx.core import MktIndirectionFactory, EnforceArchivableEntitlements

import math

__all__ = ["Currency"]


###############################################################################
@EnforceArchivableEntitlements("Database", "ArchivedOverwritable")
class Currency(Archivable):
    """
    Class used to represent a currency and its discount curve.
    """
    HolidayCalendar = ReferenceField(obj_type="HolidayCalendar")

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(CurveField)
    def ZeroCurve(self, graph):
        pass

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def DiscountFactor(self, graph, from_date, to_date):
        zero_crv = graph(self, "ZeroCurve")
        ref_date = graph("Database", "MktDataDate")
        rate_from = Interpolate(zero_crv, from_date, method="Linear")
        term_from = CalcTerm(ref_date, from_date)
        rate_to = Interpolate(zero_crv, to_date, method="Linear")
        term_to = CalcTerm(ref_date, to_date)
        return math.exp(-term_from*rate_from)*math.exp(term_to*rate_to)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import Curve, Date, AddIfMissing, EvalBlock
    from . import ufo_database
    from . import ufo_holiday_calendar

    ufo_database.prepare_for_test()
    ufo_holiday_calendar.prepare_for_test()

    ccys = []
    ccys.append(AddIfMissing(Currency(Name="USD", HolidayCalendar="USD_CAL")))
    ccys.append(AddIfMissing(Currency(Name="EUR", HolidayCalendar="EUR_CAL")))
    ccys.append(AddIfMissing(Currency(Name="GBP", HolidayCalendar="GBP_CAL")))

    zero_curve = Curve([Date.low_date()], [0.0])

    with EvalBlock() as eb:
        eb.change_value("Database", "ArchivedOverwritable", True)
        for ccy in ccys:
            ccy.set_dated("ZeroCurve", Date.today(), zero_curve)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetVal, OnyxTestCase, EvalBlock
from onyx.equity import ufo_equity_asset

import unittest


###############################################################################
class UnitTest(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        ufo_equity_asset.prepare_for_test()

    # -------------------------------------------------------------------------
    def test_Ticker(self):
        self.assertEqual(GetVal("EQ IBM US", "Ticker"), "IBM")
        self.assertEqual(GetVal("EQ NG/ LN", "Ticker"), "NG/")

        with EvalBlock() as eb:
            eb.change_value("Settings", "Platform", "Google")
            self.assertEqual(GetVal("EQ NG/ LN", "Ticker"), "NG")

        with EvalBlock() as eb:
            eb.change_value("Settings", "Platform", "Yahoo")
            self.assertEqual(GetVal("EQ NG/ LN", "Ticker"), "NG")

        with EvalBlock() as eb:
            eb.change_value("Settings", "Platform", "WSJ")
            self.assertEqual(GetVal("EQ NG/ LN", "Ticker"), "NG.")

    # -------------------------------------------------------------------------
    def test_UniqueId(self):
        self.assertEqual(GetVal("EQ IBM US", "UniqueId"), "IBM US")
        self.assertEqual(GetVal("EQ NG/ LN", "UniqueId"), "NG/ LN")

    # -------------------------------------------------------------------------
    def test_Spot(self):
        self.assertEqual(GetVal("EQ IBM US", "Spot"), 100.0)
        self.assertEqual(GetVal("EQ NG/ LN", "Spot"), 9.0)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Structure, GetObj, GetVal, InvalidateNode, EvalBlock
from onyx.core import OnyxTestCase
from onyx.system import trade_api, transaction_api

import unittest


###############################################################################
class UnitTest(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        transaction_api.prepare_for_test()
        self.cash_trade, self.cfd_trade = trade_api.prepare_for_test()

    # -------------------------------------------------------------------------
    def compareStructures(self, first, second, places):
        self.assertEqual(first.keys(), second.keys())
        for key, value in first.items():
            self.assertAlmostEqual(value, second[key], places)

    # -------------------------------------------------------------------------
    def test_MktValUSD(self):
        self.assertAlmostEqual(GetVal(self.cash_trade, "MktValUSD"), 150.0, 8)
        self.assertAlmostEqual(GetVal(self.cfd_trade, "MktValUSD"), 150.0, 8)

    # -------------------------------------------------------------------------
    def test_Deltas(self):
        ref = Structure([("EQ NG/ LN", -1000.0)])
        self.compareStructures(GetVal(self.cash_trade, "Deltas"), ref, 4)
        self.compareStructures(GetVal(self.cfd_trade, "Deltas"), ref, 4)

    # -------------------------------------------------------------------------
    def test_Exposures(self):
        usd_exposure = 13500.0
        eur_exposure = usd_exposure / 1.15
        # --- cash trade is denominated in EUR (PaymentUnit is EUR)
        self.assertAlmostEqual(eur_exposure,
                               GetVal(self.cash_trade, "GrossExposure"), 4)
        self.assertAlmostEqual(-eur_exposure,
                               GetVal(self.cash_trade, "NetExposure"), 4)
        # --- cfd trade is denominated in USD (PaymentUnit is not set)
        self.assertAlmostEqual(usd_exposure,
                               GetVal(self.cfd_trade, "GrossExposure"), 4)
        self.assertAlmostEqual(-usd_exposure,
                               GetVal(self.cfd_trade, "NetExposure"), 4)

    # -------------------------------------------------------------------------
    def test_BookAmendDelete(self):
        cash_trade = GetObj(self.cash_trade)
        cfd_trade = GetObj(self.cfd_trade)
        assert cash_trade.Party == cfd_trade.Party
        book = cash_trade.Party

        cash_trade = trade_api.CommitTrade(cash_trade)
        cfd_trade = trade_api.CommitTrade(cfd_trade)

        self.assertAlmostEqual(GetVal(book, "MktValUSD"), 300.0, 8)

        # --- trying deleting a trade without permissions should raise a
        #     TradeError exception. See below for the correct approach.
        with self.assertRaises(trade_api.TradeError):
            trade_api.DeleteTrade(cash_trade.Name)

        with EvalBlock() as eb:
            eb.change_value("Database", "TradesDeletable", True)
            trade_api.DeleteTrade(cash_trade.Name)

        # --- NB: Children needs to be invalidated manually because Book's
        #         Children are loaded directly from PosEffects table via
        #         ChildrenByBook
        InvalidateNode(book, "Children")
        self.assertAlmostEqual(GetVal(book, "MktValUSD"), 150.0, 8)

        with EvalBlock() as eb:
            eb.change_value("Database", "TradesDeletable", True)
            trade_api.DeleteTrade(cfd_trade.Name)

        # --- see comment above
        InvalidateNode(book, "Children")
        self.assertAlmostEqual(GetVal(book, "MktValUSD"), 0.0, 8)



if __name__ == "__main__":
    from onyx.core import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import UfoBase, SetField, HolidayCalendar as core_hc

__all__ = ["HolidayCalendar"]


###############################################################################
class HolidayCalendar(UfoBase, core_hc):
    """
    Class used to store holidays.
    """
    Holidays = SetField(default=set())

    # -------------------------------------------------------------------------
    @property
    def holidays(self):
        return self.Holidays


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing
    AddIfMissing(HolidayCalendar(Name="USD_CAL"))
    AddIfMissing(HolidayCalendar(Name="EUR_CAL"))
    AddIfMissing(HolidayCalendar(Name="GBP_CAL"))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.quantlib.tree_pricers import opt_BT
from onyx.quantlib.test.test_black_scholes import BLACK_SCHOLES_REF

import unittest


###############################################################################
class UnitTest(unittest.TestCase):
    # -------------------------------------------------------------------------
    def test_opt_BS(self):
        for parms, (premium, delta) in BLACK_SCHOLES_REF.items():
            # --- NB: convergence to the exact value is very slow...
            self.assertAlmostEqual(opt_BT(*parms, n_steps=510), premium, 2)
            self.assertAlmostEqual(opt_BT(*parms, n_steps=4080), premium, 3)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, RDate, Structure, ValueType, GetVal
from onyx.core import FloatField, DateField, ReferenceField

from ..system.tradable_api import AgingTradableObj
from ..system.tradable_api import AddByInference, HashStoredAttrs
from ..system.ufo_forward_cash import ForwardCash
from ..system.ufo_cfd_fixed_leg import CfdFixedLeg
from ..system.ufo_cfd_floating_leg import CfdFloatingLeg

__all__ = ["EquityCFD"]


###############################################################################
class EquityCFD(AgingTradableObj):
    """
    Tradable class that represents a contract for difference on a equity asset.
    """
    Asset = ReferenceField(obj_type="EquityAsset")
    FixedPrice = FloatField(positive=True)

    # --- the settlement currency can differ from the denomination of the asset
    SettlementCcy = ReferenceField(obj_type="Currency")

    # --- fx rate uesd to convert the fixed price (which is expressed in the
    #     currency of the asset) to the settlement currency
    SettlementFx = FloatField()

    # --- last known ex-dividend date: this is used to get a new ImpliedName
    #     every time a dividend is paid (needed by the aging mechanism).
    LastDvdDate = DateField(default=Date.low_date())

    # -------------------------------------------------------------------------
    def __post_init__(self):
        asset_ccy = GetVal(self.Asset, "Denominated")

        self.SettlementCcy = self.SettlementCcy or asset_ccy

        if self.SettlementFx is None:
            # --- default to spot
            if asset_ccy == self.SettlementCcy:
                fx = 1.0
            else:
                asset_cross = "{0:3s}/USD".format(asset_ccy)
                sett_cross = "{0:3s}/USD".format(self.SettlementCcy)
                fx = GetVal(asset_cross, "Spot") / GetVal(sett_cross, "Spot")

            self.SettlementFx = fx

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        """
        Return a Structure with security names as keys and net quantities as
        values.
        """
        asset = graph(self, "Asset")
        ccy = graph(self, "SettlementCcy")

        prc = graph(self, "FixedPrice")
        mul = graph(asset, "Multiplier")
        fx = graph(self, "SettlementFx")

        # --- leaves are created in memory
        floating = AddByInference(CfdFloatingLeg(Asset=asset), True)
        fixed = AddByInference(CfdFixedLeg(Currency=ccy), True)

        return Structure({
            floating.Name: 1.0,
            fixed.Name: -prc*mul*fx,
        })

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm = 0.0
        for sec, qty in graph(self, "Leaves").items():
            mtm += qty*graph(sec, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        sett_ccy = graph(self, "SettlementCcy")
        spot_fx = graph("{0:3s}/USD".format(sett_ccy), "Spot")
        return graph(self, "MktValUSD") / spot_fx

    # -------------------------------------------------------------------------
    @ValueType()
    def GetLastDvdDate(self, graph):
        """
        Return the last known ex-dividend date.
        """
        return graph(graph(self, "Asset"), "GetDividends").back.date

    # -------------------------------------------------------------------------
    @ValueType()
    def NextTransactionDate(self, graph):
        start = graph(self, "LastDvdDate") + RDate("1d")
        actions = graph(graph(self, "Asset"), "NextExDateActions", start)
        return actions[0]["Ex-Date"]

    # -------------------------------------------------------------------------
    @ValueType()
    def TradeTypes(self, graph):
        mapping = super().TradeTypes
        mapping.update({
            "ExDate": "ExDateSecurities",
        })
        return mapping

    # -------------------------------------------------------------------------
    @ValueType()
    def ExpectedTransaction(self, graph):
        return "ExDate"

    # -------------------------------------------------------------------------
    @ValueType()
    def ExDateSecurities(self, graph):
        ccy = graph(graph(self, "Asset"), "Denominated")
        start = graph(self, "LastDvdDate") + RDate("1d")
        ntd = graph(self, "NextTransactionDate")
        actions = graph(graph(self, "Asset"), "NextExDateActions", start)

        # --- the rolled cash-equity
        rolled = self.clone()
        rolled.LastDvdDate = ntd

        securities = [
            {"Security": AddByInference(rolled, in_memory=True),
             "Quantity": 1.0}]

        for action in actions:
            action_type = action["type"]
            if action_type == "cash":
                if ntd != action["ex-date"]:
                    raise RuntimeError("Something very wrong is going on here")
                # --- dividend receivable
                pay_date = action["pay-date"] or Date.high_date()
                pay_qty = action["amount"]
                cash = ForwardCash(Currency=ccy, PaymentDate=pay_date)

                securities += [
                    {"Security": AddByInference(cash, in_memory=True),
                     "Quantity": pay_qty}]

            elif action_type == "rights issue":
                # --- by default assume that the rights issue is on the same
                #     share class (it's not always the case)
                securities += [
                    {"Security": AddByInference(rolled, in_memory=True),
                     "Quantity": 1.0 - action["amount"]}]

            else:
                err_msg = ("Action type for {0:s} is "
                           "{1:s}").format(self.Asset, action_type)
                raise NotImplementedError(err_msg)

        return securities

    # -------------------------------------------------------------------------
    #  this value type is used by edit screens
    @ValueType()
    def AssetSpot(self, graph):
        return graph(graph(self, "Asset"), "Spot")

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        sym = GetVal(self.Asset, "Symbol")
        code = GetVal(GetVal(self.Asset, "Exchange"), "Code")
        mush = HashStoredAttrs(self, 4)
        return ("EqCFD {0:s} {1:2s} {2:3s} {3:4s} "
                "{{0:2d}}").format(sym, code, self.SettlementCcy, mush)


###############################################################################
class EquityCFDCalc(EquityCFD):
    """
    Calculator class.
    """
    PriceAdjust = FloatField(default=0.0)

    # -------------------------------------------------------------------------
    @ValueType()
    def TradableFactory(self, graph):
        """
        Return a list of tradables objects that replicate the calculator.
        """
        info = {"Asset": graph(self, "Asset"),
                "FixedPrice": graph(self, "FixedPrice"),
                "SettlementCcy": graph(self, "SettlementCcy"),
                "SettlementFx": graph(self, "SettlementFx")}
        return [EquityCFD(**info)]

    # -------------------------------------------------------------------------
    @ValueType()
    def AssetSpot(self, graph):
        return graph(self, "PriceAdjust") + graph(graph(self, "Asset"), "Spot")

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        dnpv = 0.0
        for sec, qty in graph(self, "Leaves").items():
            dnpv += qty*graph(sec, "MktValUSD")
        return dnpv + graph(self, "PriceAdjust")

    # -------------------------------------------------------------------------
    def spot_settlement_fx(self):
        ccy1 = GetVal(GetVal(self, "Asset"), "Denominated")
        ccy2 = GetVal(self, "SettlementCcy")
        cross1 = "{0:3s}/USD".format(ccy1)
        cross2 = "{0:3s}/USD".format(ccy2)
        fx_rate = GetVal(cross1, "Spot") / GetVal(cross2, "Spot")

        return fx_rate


# -----------------------------------------------------------------------------
def prepare_for_test():
    from ..system import ufo_forward_cash
    from . import ufo_equity_asset

    ufo_equity_asset.prepare_for_test()
    ufo_forward_cash.prepare_for_test()

    info = {
        "Asset": "EQ NG/ LN",
        "FixedPrice": 910.0,
    }

    securities = [AddByInference(EquityCFD(**info))]

    return [sec.Name for sec in securities]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetVal, ValueType
from onyx.core import IntField, DateField
from onyx.core import SelectField, BoolField, ReferenceField

from .ufo_commod_nrby import AVERAGING_TYPES, CommodNrby

from ..system.tradable_api import NamedByInference, AddByInference
from ..system.tradable_api import HashStoredAttrs

__all__ = ["CommodForward"]


###############################################################################
class CommodForward(NamedByInference):
    """
    class that represents the floating leg of a commodity forward or futures
    contract that is settled in cash.
    """
    Asset = ReferenceField(obj_type="CommodAsset")
    AvgStartDate = DateField()
    AvgEndDate = DateField()
    AvgType = SelectField(options=AVERAGING_TYPES)
    RollType = IntField(default=0)
    Margined = BoolField(default=False)
    Denominated = ReferenceField(obj_type="Currency")

    # -------------------------------------------------------------------------
    def __post_init__(self):
        # --- set defaults for attributes inherited from Asset
        for attr in ["Denominated"]:
            if getattr(self, attr) is None:
                setattr(self, attr, GetVal(self.Asset, attr))

    # -------------------------------------------------------------------------
    @ValueType()
    def Nearby(self, graph):
        # --- create the nearby object in memory
        info = {
            "Asset": graph(self, "Asset"),
            "StartDate": graph(self, "AvgStartDate"),
            "EndDate": graph(self, "AvgEndDate"),
            "AvgType": graph(self, "AvgType"),
            "RollType": graph(self, "RollType"),
        }
        nrby = AddByInference(CommodNrby(**info), in_memory=True)
        return graph(nrby, "Name")

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        avg_val = graph(graph(self, "Nearby"), "AverageValue")
        if graph(self, "Margined"):
            df = 1.0
        else:
            pd = graph("Database", "PricingDate")
            ed = graph(self, "ExpirationDate")
            ccy = graph(self, "Denominated")
            df = graph(ccy, "DiscountFactor", ed, pd)
        return df*avg_val

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        fx = graph("{0:3s}/USD".format(graph(self, "Denominated")), "Spot")
        return fx*graph(self, "MktVal")

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        mkt = GetVal(self.Asset, "Market")
        sym = GetVal(self.Asset, "Symbol")
        mush = HashStoredAttrs(self, 8)
        return "CmdFWD {0:s} {1:s} {2:8s} {{0:2d}}".format(mkt, sym, mush)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################
"""
Scheduling functions and request handlers implementing a RESTful API.
"""
from apscheduler.schedulers.tornado import TornadoScheduler

import apscheduler.jobstores.base
import tornado.web
import importlib
import json
import shlex
import logging

__all__ = [
    "SchedulerError",
    "IndexHandler",
    "JobsHandler",
    "SchedulingServer"
]

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
def run_job(mod_name, args, kwds):
    module = importlib.import_module(mod_name)
    module.run(*args, **kwds)


# -----------------------------------------------------------------------------
def arg_split(arg):
    key, value = arg.split("=")
    # --- cast boolean strings to proper datatype
    if value.upper() == "TRUE":
        value = True
    elif value.upper() == "FALSE":
        value = False
    return key.lstrip("-"), value


# -----------------------------------------------------------------------------
def arg_combine(option, value):
    return "--{0:s}={1!s}".format(option, value)


###############################################################################
class SchedulerError(Exception):
    pass


###############################################################################
class IndexHandler(tornado.web.RequestHandler):
    # -------------------------------------------------------------------------
    def initialize(self, index_file):
        self.index_file = index_file

    # -------------------------------------------------------------------------
    def get(self):
        with open(self.index_file, "r") as index_file:
            self.write(index_file.read())


###############################################################################
class JobsHandler(tornado.web.RequestHandler):
    # -------------------------------------------------------------------------
    def get(self, job_id=None):
        logger.debug("GET: job_id is {0!s}".format(job_id))

        if job_id is None:
            # --- return all pending jobs
            response = []
            for job in self.application.scheduler.get_jobs():
                if job.next_run_time is None:
                    next_run_time = "paused"
                else:
                    next_run_time = job.next_run_time.isoformat()

                response.append({
                    "name": job.name,
                    "kwds": " ".join([arg_combine(k, v)
                                      for k, v in job.kwargs["kwds"].items()]),
                    "id": job.id,
                    "executor": job.executor,
                    "next_run_time": next_run_time,
                    "trigger": {field.name: str(field)
                                for field in job.trigger.fields},
                })

        else:
            response = "Not implemented...\n"

        self.set_status(200)
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        self.write(json.dumps(response))

    # -------------------------------------------------------------------------
    def put(self, job_id=None):
        logger.debug("PUT: job_id is {0!s}".format(job_id))
        self.set_header("Access-Control-Allow-Origin", "*")

        if job_id is None:
            self.set_status(404)
            return

        changes = json.loads(self.request.body.decode("utf-8"))
        trigger_args = changes.pop("trigger", None)
        job_status = changes.pop("next_run_time", None)

        try:
            self.application.scheduler.modify_job(job_id, **changes)
            if trigger_args is not None:
                self.application.scheduler.reschedule_job(job_id,
                                                          trigger="cron",
                                                          **trigger_args)
                logger.info("PUT: job was rescheduled")

            if job_status == "pause":
                self.application.paused.add(job_id)
                self.application.scheduler.pause_job(job_id)
                logger.info("PUT: job was paused")

            elif job_status == "resume":
                if job_id in self.application.paused:
                    self.application.scheduler.resume_job(job_id)
                    self.application.paused.remove(job_id)
                    logger.info("PUT: job was resumed")

        except apscheduler.jobstores.base.JobLookupError as err:
            self.set_status(404)
            self.write("{0!s}\n".format(err))

    # -------------------------------------------------------------------------
    def delete(self, job_id=None):
        logger.debug("DELETE: job_id is {0!s}".format(job_id))
        self.set_header("Access-Control-Allow-Origin", "*")
        if job_id is None:
            self.set_status(404)
            return

        try:
            self.application.scheduler.remove_job(job_id)
        except apscheduler.jobstores.base.JobLookupError as err:
            self.set_status(404)
            self.write("{0!s}\n".format(err))

    # -------------------------------------------------------------------------
    def post(self, job_id=None):
        logger.debug("POST: job_id is {0!s}".format(job_id))
        self.set_header("Access-Control-Allow-Origin", "*")

        if job_id is not None:
            self.set_status(404)
            return

        job_parms = json.loads(self.request.body.decode("utf-8"))
        try:
            self.application.schedule_job({
                "module": job_parms["name"],
                "trigger": job_parms["trigger"],
                "kwds": dict([arg_split(arg)
                              for arg in shlex.split(job_parms["kwds"])]),
            })
            self.set_status(201)
        except SchedulerError as err:
            self.set_status(400)
            self.write("{0!s}\n".format(err))

    # -------------------------------------------------------------------------
    def options(self, job_id=None):
        logger.debug("OPTIONS: job_id is {0!s}".format(job_id))
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "GET,PUT,POST,DELETE")
        self.set_header("Access-Control-Allow-Headers",
                        "Content-Type,Authorization,accept,x-requested-with")
        self.set_status(200)


###############################################################################
class SchedulingServer(tornado.web.Application):
    # -------------------------------------------------------------------------
    def __init__(self, jobstores, executors, handlers=None, **kwds):
        super().__init__(handlers, **kwds)

        self.scheduler = TornadoScheduler(logger=logger,
                                          jobstores=jobstores,
                                          executors=executors)
        self.scheduler.start()

        self.paused = set()
        for job in self.scheduler.get_jobs():
            if job.next_run_time is None:
                self.paused.add(job.id)

    # -------------------------------------------------------------------------
    def schedule_job(self, job):
        mod_name = job["module"]

        # --- is the module importable and does it expose a run function?
        try:
            module = importlib.import_module(mod_name)
        except (ImportError, SystemError) as err:
            raise SchedulerError("{0!s}".format(err))

        if not hasattr(module, "run"):
            raise SchedulerError("{0:s} doesn't expose a run "
                                 "function".format(job["module"]))

        kwargs = {
            "mod_name": mod_name,
            "args": job.get("args", []),
            "kwds": job.get("kwds", {}),
        }

        # --- each job is scheduled using a wrapper
        return self.scheduler.add_job(run_job, kwargs=kwargs,
                                      trigger="cron",
                                      name=job["module"],
                                      replace_existing=True, **job["trigger"])
